# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableStrengthImportationMixin,
    EnableVolumeImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class BinarySvipVersion(Enum):
    AUTO: Annotated[
        str,
        Field(
            title=_("Auto detect"),
            description=_(
                "Use the version of the project file corresponding to the X Studio installed on your current system."
            ),
        ),
    ] = "auto"
    SVIP7_0_0: Annotated[
        str,
        Field(
            title="SVIP 7.0.0",
            description=_("Use the project file version of X Studio 2.0."),
        ),
    ] = "7.0.0"
    SVIP6_0_0: Annotated[
        str,
        Field(
            title="SVIP 6.0.0",
            description=_("Use the project file version which is compatible with X Studio 1.8."),
        ),
    ] = "6.0.0"
    COMPAT: Annotated[
        str,
        Field(
            title=_("Max compatibility (read only)"),
            description=_("""Export project files that can be opened using any version of X Studio.
Warning: After saving with this option, the volume, breath, gender, and power parameters will not be recognized by X Studio (no data is lost).
To avoid irretrievable data loss, it is strongly recommended not to use X Studio to modify and save project files exported with this option. To restore the project file back to a safe editable state, select Save as SVIP 6.0.0 or later."""),
        ),
    ] = "0.0.0"


class InputOptions(
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableStrengthImportationMixin,
    EnableVolumeImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    singer: str = Field(
        default="陈水若",
        title=_("Default singer"),
        description=_(
            "Please enter the singer's name in Chinese. If the specified singer does not exist, the default singer set in X Studio will be used. If you want to specify the conversion relationship between singer ID and name, or add a singer that has an ID but has not been publicly released, please modify singers.json in the plugin directory."
        ),
    )
    tempo: int = Field(
        default=60,
        title=_("Default tempo"),
        description=_(
            "The allowed range of tempo in X Studio is 20 ~ 300. When the tempo is out of range, the absolute timeline will be used to ensure the alignment of notes. Please set this value to an integer multiple or integer fraction of the tempo in the source project file as much as possible; as long as it is within a reasonable range, the value of this option will not affect the alignment effect."
        ),
    )
    version: BinarySvipVersion = Field(
        default=BinarySvipVersion.AUTO,
        title=_("Specify the version of the generated .svip file"),
        description=_("""This option only controls the header version information of the output project file.
Choosing an older project file version will not affect the integrity of the data, but using a lower version of X Studio to open, edit and save a higher version project file may cause data loss."""),
    )



import re
from base64 import b64encode
from hashlib import sha224

from yaml.resolver import BaseResolver

merge_resolver = {
    "tag": "tag:yaml.org,2002:merge",
    "regexp": re.compile(r"^(?:<<)$"),
    "start_chars": ["<"],
}
core_resolvers = {
    "1.2": [
        {
            "tag": "tag:yaml.org,2002:bool",
            "regexp": re.compile(r"^(?:|true|True|TRUE|false|False|FALSE)$", re.VERBOSE),
            "start_chars": list("tTfF"),
        },
        {
            "tag": "tag:yaml.org,2002:int",
            "regexp": re.compile(r"^(?:|0o[0-7]+|[-+]?(?:[0-9]+)|0x[0-9a-fA-F]+)$", re.VERBOSE),
            "start_chars": list("-+0123456789"),
        },
        {
            "tag": "tag:yaml.org,2002:float",
            "regexp": re.compile(
                r"^(?:[-+]?(?:\.[0-9]+|[0-9]+(\.[0-9]*)?)(?:[eE][-+]?[0-9]+)?|[-+]?\.(?:inf|Inf|INF)|\.(?:nan|NaN|NAN))$",
                re.VERBOSE,
            ),
            "start_chars": list("-+0123456789."),
        },
        {
            "tag": "tag:yaml.org,2002:null",
            "regexp": re.compile(r"^(?:~||null|Null|NULL)$", re.VERBOSE),
            "start_chars": ["~", "n", "N", ""],
        },
    ],
}


def set_yaml_grammar(
    resolver: BaseResolver,
    grammar_version: str = "1.2",
    expand_merge_keys: bool = True,
) -> None:
    if grammar_version not in core_resolvers:
        msg = f"Unknown grammar version {grammar_version}"
        raise ValueError(msg)
    resolvers = list(core_resolvers[grammar_version])
    if expand_merge_keys:
        resolvers.append(merge_resolver)
    resolver.yaml_implicit_resolvers = {}
    for r in resolvers:
        for start_char in r["start_chars"]:  # type: ignore[attr-defined]
            resolver.yaml_implicit_resolvers.setdefault(start_char, [])
            resolver.yaml_implicit_resolvers[start_char].append((r["tag"], r["regexp"]))


def hash_key(key: str | bytes) -> str:
    return b64encode(sha224(key.encode() if isinstance(key, str) else key).digest()).decode()



__all__ = [
    "Clipboard",
    "ConfigItems",
    "FramelessWindow",
    "IconicFontLoader",
    "LocaleSwitcher",
    "TaskManager",
    "app",
    "app_close_event",
    "event_loop",
    "qml_engine",
]

import platform
import sys

from .application import app, app_close_event, event_loop, qml_engine
from .clipboard import Clipboard
from .config_items import ConfigItems
from .font_loader import IconicFontLoader
from .locale_switcher import LocaleSwitcher
from .log_handler import enable_log_handler
from .task_manager import TaskManager

if sys.platform == "win32":
    from libresvip.gui.modules.frameless_window_win32 import FramelessWindow
else:
    from libresvip.gui.modules.frameless_window import FramelessWindow

enable_log_handler()
if platform.python_implementation() == "CPython":
    __all__ += ["Notifier"]

    from .notifier import Notifier



# Yet another MS-NRBF parser and serializer for Python
# Special thanks: netfleece, pypdn, https://github.com/gurnec/Undo_FFG
import abc
import decimal
import threading
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from functools import partial
from typing import Any, BinaryIO

from construct import (
    Adapter,
    BitsInteger,
    BitStruct,
    Byte,
    BytesInteger,
    Computed,
    Const,
    Construct,
    Container,
    ExprAdapter,
    Flag,
    Float32l,
    Float64l,
    If,
    IfThenElse,
    Int8sl,
    Int8ul,
    Int16sl,
    Int16ul,
    Int64sl,
    Int64ul,
    LazyBound,
    PascalString,
    PrefixedArray,
    RepeatUntil,
    SizeofError,
    Struct,
    Switch,
    this,
)
from construct import Enum as CSEnum
from construct import Path as CSPath
from construct_typed import Context
from typing_extensions import Never

Int32ul = BytesInteger(4, swapped=True)
Int32sl = BytesInteger(4, swapped=True, signed=True)


class TimeSpanAdapter(Adapter):
    def _encode(self, obj: timedelta, context: Context, path: CSPath) -> int:
        return int(obj.total_seconds() * 10000000)

    def _decode(self, obj: int, context: Context, path: CSPath) -> timedelta:
        return timedelta(microseconds=obj / 10)


TimeSpan = TimeSpanAdapter(Int64ul)

DateTimeBitStruct = BitStruct(
    "ticks" / BitsInteger(62),
    "kind" / BitsInteger(2),
)


class DateTimeAdapter(Adapter):
    def _encode(self, obj: datetime, context: Context, path: CSPath) -> bytes:
        if obj.tzinfo is None:
            kind = 0
        elif obj.tzinfo.utcoffset(obj) == timedelta(0):
            kind = 1
        else:
            kind = 2
        ticks = (obj - datetime(1, 1, 1, tzinfo=timezone.utc)).total_seconds() * 10000000
        return DateTimeBitStruct.build({"ticks": ticks, "kind": kind})

    def _decode(self, obj: Container, context: Context, path: CSPath) -> datetime:
        date_time = datetime(1, 1, 1, tzinfo=timezone.utc) + timedelta(microseconds=obj.ticks / 10)
        if obj.kind == 1:
            date_time = date_time.replace(tzinfo=timezone.utc)
        elif obj.kind == 2:
            local_timezone = datetime.now(tz=timezone.utc).astimezone().tzinfo
            date_time = date_time.replace(tzinfo=local_timezone)
        return date_time


DateTime = DateTimeAdapter(DateTimeBitStruct)


class Null(Construct):
    def _sizeof(self, context: Context, path: CSPath) -> int:
        return 0

    def _parse(self, stream: BinaryIO, context: Context, path: CSPath) -> None:
        return None

    def _build(self, obj: Container, stream: BinaryIO, context: Context, path: CSPath) -> None:
        pass


class Utf8CodePoint(Construct):
    def _sizeof(self, context: Context, path: CSPath) -> Never:
        msg = "Utf8CodePoint has no static size"
        raise SizeofError(msg)

    def _parse(self, stream: BinaryIO, context: Context, path: CSPath) -> str:
        byte = stream.read(1)
        if not byte:
            raise EOFError
        b = ord(byte)
        if b & 0x80 == 0:
            length = 1
        elif b & 0xE0 == 0xC0:
            length = 2
        elif b & 0xF0 == 0xE0:
            length = 3
        elif b & 0xF8 == 0xF0:
            length = 4
        else:
            msg = "Invalid UTF-8 code point"
            raise ValueError(msg)
        return (byte + stream.read(length - 1)).decode("utf-8")

    def _build(
        self,
        obj: str | bytes,
        stream: BinaryIO,
        context: Context,
        path: CSPath,
    ) -> bytes:
        if isinstance(obj, str):
            obj = obj.encode("utf-8")
        stream.write(obj)
        return obj


class LengthPrefixedString(Construct):
    def _sizeof(self, context: Context, path: CSPath) -> Never:
        msg = "LengthPrefixedString has no static size"
        raise SizeofError(msg)

    def _parse(self, stream: BinaryIO, context: Context, path: CSPath) -> str:
        length = 0
        shift = 0
        for i in range(5):
            byte = stream.read(1)
            if not byte:
                raise EOFError
            b = ord(byte)
            length += (b & ~0x80) << shift
            shift += 7
            if not b & 0x80:
                break
        else:
            msg = "Invalid length-prefixed string"
            raise ValueError(msg)
        content = stream.read(length)
        return content.decode("utf-8")

    def _build(
        self,
        obj: str | bytes,
        stream: BinaryIO,
        context: Context,
        path: CSPath,
    ) -> bytes:
        if isinstance(obj, str):
            obj = obj.encode("utf-8")
        length = len(obj)
        while length > 0x7F:
            stream.write(bytes(((length & 0x7F) | 0x80,)))
            length >>= 7
        stream.write(bytes((length,)))
        stream.write(obj)
        return obj


Decimal = ExprAdapter(
    LengthPrefixedString(),
    lambda obj, ctx: decimal.Decimal(obj),
    lambda obj, ctx: str(obj),
)


classes_by_id: dict[int, dict[int, Container]] = defaultdict(dict)
objects_by_id: dict[int, dict[int, Container]] = defaultdict(dict)
libraries_by_id: dict[int, dict[int, str]] = defaultdict(dict)
references_by_id: dict[int, dict[int, Container]] = defaultdict(dict)


class RegistryAdapter(Adapter, abc.ABC):
    def _encode(self, obj: Container, context: Context, path: CSPath) -> Any:
        return obj


class ClassRegistryAdapter(RegistryAdapter):
    def _decode(self, obj: Container, context: Context, path: CSPath) -> Any:
        classes_by_id[threading.get_ident()][obj.class_info.object_id] = obj
        return obj


class ObjectRegistryAdapter(RegistryAdapter):
    def _decode(self, obj: Container, context: Context, path: CSPath) -> Any:
        if obj.get("array_info", None):
            objects_by_id[threading.get_ident()][obj.array_info.object_id] = obj
        else:
            objects_by_id[threading.get_ident()][obj.object_id] = obj
        return obj


class LibraryRegistryAdapter(RegistryAdapter):
    def _decode(self, obj: Container, context: Context, path: CSPath) -> Any:
        libraries_by_id[threading.get_ident()][obj.library_id] = obj.library_name
        return obj


class MemberReferenceAdapter(RegistryAdapter):
    def _decode(self, obj: Container, context: Context, path: CSPath) -> Any:
        ref_cache = references_by_id[threading.get_ident()]
        if obj.id_ref not in ref_cache:
            result = {"id_ref": obj.id_ref, "real_obj": None}
            ref_cache[obj.id_ref] = result
        return ref_cache[obj.id_ref]


PrimitiveTypeEnum = CSEnum(
    Byte,
    Boolean=1,
    Byte=2,
    Char=3,
    Decimal=5,
    Double=6,
    Int16=7,
    Int32=8,
    Int64=9,
    SByte=10,
    Single=11,
    TimeSpan=12,
    DateTime=13,
    UInt16=14,
    UInt32=15,
    UInt64=16,
    Null=17,
    String=18,
)

RecordTypeEnum = CSEnum(
    Byte,
    SerializedStreamHeader=0,
    ClassWithId=1,
    SystemClassWithMembers=2,
    ClassWithMembers=3,
    SystemClassWithMembersAndTypes=4,
    ClassWithMembersAndTypes=5,
    BinaryObjectString=6,
    BinaryArray=7,
    MemberPrimitiveTyped=8,
    MemberReference=9,
    ObjectNull=10,
    MessageEnd=11,
    BinaryLibrary=12,
    ObjectNullMultiple256=13,
    ObjectNullMultiple=14,
    ArraySinglePrimitive=15,
    ArraySingleObject=16,
    ArraySingleString=17,
    MethodCall=21,
    MethodReturn=22,
)

BinaryTypeEnum = CSEnum(
    Byte,
    Primitive=0,
    String=1,
    Object=2,
    SystemClass=3,
    Class=4,
    ObjectArray=5,
    StringArray=6,
    PrimitiveArray=7,
)

BinaryArrayTypeEnum = CSEnum(
    Byte,
    Single=0,
    Jagged=1,
    Rectangular=2,
    SingleOffset=3,
    JaggedOffset=4,
    RectangularOffset=5,
)

MessageFlagsEnum = CSEnum(
    Int32ul,
    NoArgs=0x00000001,
    ArgsInline=0x00000002,
    ArgsIsArray=0x00000004,
    ArgsInArray=0x00000008,
    NoContext=0x00000010,
    ContextInline=0x00000020,
    ContextInArray=0x00000040,
    MethodSignatureInArray=0x00000080,
    PropertiesInArray=0x00000100,
    NoReturnValue=0x00000200,
    ReturnValueVoid=0x00000400,
    ReturnValueInline=0x00000800,
    ReturnValueInArray=0x00001000,
    ExceptionInArray=0x00002000,
    GenericMethod=0x00008000,
)

PrimitiveType = partial(
    Switch,
    cases={
        "Boolean": Flag,
        "Byte": Int8ul,
        "Char": Utf8CodePoint,
        "Decimal": Decimal,
        "Double": Float64l,
        "Int16": Int16sl,
        "Int32": Int32sl,
        "Int64": Int64sl,
        "SByte": Int8sl,
        "Single": Float32l,
        "TimeSpan": TimeSpan,
        "DateTime": DateTime,
        "UInt16": Int16ul,
        "UInt32": Int32ul,
        "UInt64": Int64ul,
        "Null": Null(),
        "String": LengthPrefixedString(),
    },
)

ArrayInfo = Struct(
    "object_id" / Int32sl,
    "length" / Int32sl,
)

ClassTypeInfo = Struct(
    "type_name" / LengthPrefixedString(),
    "library_id" / Int32sl,
)

ClassInfo = Struct(
    "object_id" / Int32sl,
    "name" / LengthPrefixedString(),
    "member_count" / Int32sl,
    "member_names" / LengthPrefixedString()[this.member_count],
)

SerializedStreamHeader = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.SerializedStreamHeader),
    "root_id" / Int32sl,
    "header_id" / Int32sl,
    "major_version" / Int32sl,
    "minor_version" / Int32sl,
)

MemberValue = Struct(
    "value" / LazyBound(lambda: Record),
)

SystemClassWithMembers = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.SystemClassWithMembers),
    "class_info" / ClassInfo,
    "member_values" / MemberValue[this.class_info.member_count],
)

ClassWithMembers = ClassRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ClassWithMembers),
        "class_info" / ClassInfo,
        "library_id" / Int32sl,
        "member_values" / MemberValue[this.class_info.member_count],
    )
)


BinaryType = partial(
    Switch,
    cases={
        "Primitive": PrimitiveTypeEnum,
        "String": Null(),
        "Object": Null(),
        "SystemClass": LengthPrefixedString(),
        "Class": ClassTypeInfo,
        "ObjectArray": Null(),
        "StringArray": Null(),
        "PrimitiveArray": PrimitiveTypeEnum,
    },
)


MemberTypeAddtionalInfo = Struct(
    "binary_type_enum" / Computed(lambda this: this._.binary_type_enums[this._index]),
    "info" / BinaryType(lambda this: this.binary_type_enum),
)


MemberTypeInfo = Struct(
    "binary_type_enums" / BinaryTypeEnum[this._.class_info.member_count],
    "additional_infos" / MemberTypeAddtionalInfo[this._.class_info.member_count],
)


MemberValueWithType = Struct(
    "binary_type_enum"
    / Computed(lambda this: (this._.member_type_info["binary_type_enums"][this._._index])),
    "value"
    / Switch(
        lambda this: this.binary_type_enum,
        {
            "Primitive": PrimitiveType(
                lambda this: this._.member_type_info["additional_infos"][this._._index]["info"]
            ),
        },
        default=LazyBound(lambda: Record),
    ),
)


SystemClassWithMembersAndTypes = ClassRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.SystemClassWithMembersAndTypes),
        "class_info" / ClassInfo,
        "member_type_info" / MemberTypeInfo,
        "member_values" / MemberValueWithType[this.class_info.member_count],
    )
)

ClassWithMembersAndTypes = ClassRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ClassWithMembersAndTypes),
        "class_info" / ClassInfo,
        "member_type_info" / MemberTypeInfo,
        "library_id" / Int32sl,
        "member_values" / MemberValueWithType[this.class_info.member_count],
    )
)

ClassWithId = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ClassWithId),
        "object_id" / Int32sl,
        "metadata_id" / Int32sl,
        "class_info"
        / Computed(
            lambda this: (classes_by_id[threading.get_ident()][this.metadata_id]["class_info"])
        ),
        "member_type_info"
        / Computed(
            lambda this: (
                classes_by_id[threading.get_ident()][this.metadata_id].get("member_type_info", None)
            )
        ),
        "member_values"
        / IfThenElse(
            lambda this: this.member_type_info,
            MemberValueWithType[this.class_info.member_count],
            MemberValue[this.class_info.member_count],
        ),
    )
)

BinaryObjectString = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.BinaryObjectString),
        "object_id" / Int32sl,
        "value" / LengthPrefixedString(),
    )
)

BinaryArray = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.BinaryArray),
        "object_id" / Int32sl,
        "binary_array_type_enum" / BinaryArrayTypeEnum,
        "rank" / Int32sl,
        "lengths" / Int32sl[this.rank],
        "lower_bounds"
        / IfThenElse(
            lambda this: str(this.binary_array_type_enum)
            in [
                "SingleOffset",
                "JaggedOffset",
                "RectangularOffset",
            ],
            Int32sl[this.rank],
            Null(),
        ),
        "binary_type_enum" / BinaryTypeEnum,
        "info" / BinaryType(lambda this: (this.binary_type_enum)),
        "member_values"
        / If(
            lambda this: (
                str(this.binary_array_type_enum)
                not in [
                    "Rectangular",
                    "RectangularOffset",
                ]
                and this.rank == 1
                and this.lengths[0] > 0
            ),
            RepeatUntil(
                lambda obj, lst, ctx: (
                    (
                        len(lst)
                        + sum(
                            (x.real_obj.obj.null_count - 1)
                            for x in lst
                            if "obj" in x.real_obj and "null_count" in x.real_obj.obj
                        )
                    )
                    >= obj["total"]
                ),
                Struct(
                    "total" / Computed(lambda this: (this._.lengths[0])),
                    "real_obj"
                    / Switch(
                        lambda this: this._.binary_type_enum,
                        {
                            "Primitive": PrimitiveType(
                                lambda this: this._._.member_type_info.additional_infos[
                                    this._._._index
                                ].info
                            ),
                            "String": LengthPrefixedString(),
                        },
                        default=LazyBound(lambda: Record),
                    ),
                ),
            ),
        ),
        # TODO: implement multidimensional arrays
    )
)

ValueWithCode = Struct(
    "primitive_type_enum" / PrimitiveTypeEnum,
    "value" / PrimitiveType(this.primitive_type_enum),
)

StringValueWithCode = Struct(
    "primitive_type_enum" / Const(PrimitiveTypeEnum.String, Byte),
    "value" / LengthPrefixedString(),
)

ArrayOfValueWithCode = PrefixedArray(Int32sl, ValueWithCode)

MemberPrimitiveTyped = (
    "record_type_enum" / Computed(RecordTypeEnum.MemberPrimitiveTyped),
    "value" / ValueWithCode,
)

MemberReference = MemberReferenceAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.MemberReference),
        "id_ref" / Int32sl,
    )
)

BinaryLibrary = LibraryRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.BinaryLibrary),
        "library_id" / Int32sl,
        "library_name" / LengthPrefixedString(),
    )
)

ObjectNullMultiple256 = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.ObjectNullMultiple256),
    "null_count" / Int8ul,
)

ObjectNullMultiple = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.ObjectNullMultiple),
    "null_count" / Int32sl,
)

ArraySinglePrimitive = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ArraySinglePrimitive),
        "array_info" / ArrayInfo,
        "primitive_type_enum" / PrimitiveTypeEnum,
        "member_values" / PrimitiveType(this.primitive_type_enum)[this.array_info.length],
    )
)

ArraySingleObject = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ArraySingleObject),
        "array_info" / ArrayInfo,
        "member_values" / LazyBound(lambda: Record[this.array_info.length]),
    )
)

ArraySingleString = ObjectRegistryAdapter(
    Struct(
        "record_type_enum" / Computed(RecordTypeEnum.ArraySingleString),
        "array_info" / ArrayInfo,
        "member_values" / LazyBound(lambda: Record[this.array_info.length]),
    )
)

MethodCall = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.MethodCall),
    "flags" / MessageFlagsEnum,
    "method_name" / StringValueWithCode,
    "type_name" / StringValueWithCode,
    "call_context"
    / If(
        this.flags & MessageFlagsEnum.ContextInline,
        StringValueWithCode,
    ),
    "args"
    / If(
        this.flags & MessageFlagsEnum.ArgsInline,
        ArrayOfValueWithCode,
    ),
)

MethodReturn = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.MethodReturn),
    "flags" / MessageFlagsEnum,
    "return_value"
    / If(
        this.flags & MessageFlagsEnum.ReturnValueInline,
        ValueWithCode,
    ),
    "call_context"
    / If(
        this.flags & MessageFlagsEnum.ContextInline,
        StringValueWithCode,
    ),
    "args"
    / If(
        this.flags & MessageFlagsEnum.ArgsInline,
        ArrayOfValueWithCode,
    ),
)


ObjectNull = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.ObjectNull),
)

MessageEnd = Struct(
    "record_type_enum" / Computed(RecordTypeEnum.MessageEnd),
)


Record: Container = Struct(
    "record_type_enum" / RecordTypeEnum,
    "obj"
    / Switch(
        this.record_type_enum,
        {
            "SerializedStreamHeader": SerializedStreamHeader,
            "ClassWithId": ClassWithId,
            "SystemClassWithMembers": SystemClassWithMembers,
            "ClassWithMembers": ClassWithMembers,
            "SystemClassWithMembersAndTypes": SystemClassWithMembersAndTypes,
            "ClassWithMembersAndTypes": ClassWithMembersAndTypes,
            "BinaryObjectString": BinaryObjectString,
            "BinaryArray": BinaryArray,
            "MemberPrimitiveTyped": MemberPrimitiveTyped,
            "MemberReference": MemberReference,
            "ObjectNull": ObjectNull,
            "MessageEnd": MessageEnd,
            "BinaryLibrary": BinaryLibrary,
            "ObjectNullMultiple256": ObjectNullMultiple256,
            "ObjectNullMultiple": ObjectNullMultiple,
            "ArraySinglePrimitive": ArraySinglePrimitive,
            "ArraySingleObject": ArraySingleObject,
            "ArraySingleString": ArraySingleString,
            "MethodCall": MethodCall,
            "MethodReturn": MethodReturn,
        },
    ),
)

RecordStream = RepeatUntil(
    lambda obj, lst, ctx: obj is None or obj["record_type_enum"] == RecordTypeEnum.MessageEnd,
    Record,
)


SVIPFile = Struct(
    "magic" / PascalString(Byte, "utf-8"),
    "version" / PascalString(Byte, "utf-8"),
    "record_stream" / RecordStream,
)



import dataclasses

from drawsvg import Drawing

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve, Project, SingingTrack

from .coordinate_helper import CoordinateHelper
from .options import OutputOptions
from .svg_factory import SvgFactory


@dataclasses.dataclass
class SvgGenerator:
    options: OutputOptions
    coordinate_helper: CoordinateHelper = dataclasses.field(init=False)
    svg_factory: SvgFactory = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> Drawing:
        self.coordinate_helper = CoordinateHelper(
            options=self.options,
            pitch_position_offset=int(project.time_signature_list[0].bar_length()),
        )
        self.svg_factory = SvgFactory(
            coordinate_helper=self.coordinate_helper,
            time_synchronizer=TimeSynchronizer(project.song_tempo_list),
            options=self.options,
        )
        if self.options.track_index < 0:
            first_singing_track = next(
                (track for track in project.track_list if isinstance(track, SingingTrack)),
                None,
            )
        else:
            first_singing_track = project.track_list[self.options.track_index]
        if isinstance(first_singing_track, SingingTrack):
            self.coordinate_helper.calculate_range(first_singing_track)
            if self.options.show_grid:
                self.svg_factory.draw_grid(project.time_signature_list)
            self.generate_notes(first_singing_track.note_list)
            self.generate_pitch(first_singing_track.edited_params.pitch)

        drawing = self.generate_svg()
        return drawing

    def generate_svg(self) -> Drawing:
        drawing = Drawing(*self.coordinate_helper.size)
        drawing.append_css(self.svg_factory.style)
        drawing.extend(self.svg_factory.line_elements)
        drawing.extend(self.svg_factory.rect_elements)
        drawing.extend(self.svg_factory.polyline_elements)
        drawing.extend(self.svg_factory.text_elements)
        return drawing

    def generate_notes(self, note_list: list[Note]) -> None:
        for note in note_list:
            self.svg_factory.draw_note(note)

    def generate_pitch(self, pitch: ParamCurve) -> None:
        for point in pitch.points.root[1:-1]:
            self.svg_factory.draw_pitch(point)



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _

from .models.enums import VocaloidLanguage, VsqxVersion


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    BaseModel,
):
    combine_syllables: bool = Field(
        False,
        title=_("Combine syllables"),
        description=_("Combine multisyllabic lyrics into single words"),
    )


class OutputOptions(BaseModel):
    vsqx_version: VsqxVersion = Field(VsqxVersion.VSQ4, title=_("VSQX Version"))
    pretty_xml: bool = Field(
        True,
        title=_("Pretty XML"),
        description=_("Whether to output pretty XML"),
    )
    default_lang_id: VocaloidLanguage = Field(
        VocaloidLanguage.SIMPLIFIED_CHINESE,
        title=_("Default language"),
        description=_("Default language id of voicebank and notes"),
    )
    default_comp_id: str = Field(
        "BETDB8W6KWZPYEB9",
        title=_("Default Comp ID"),
        description=_("Default comp_id of voicebank"),
    )
    default_singer_name: str = Field(
        "Tianyi_CHN",
        title=_("Default Singer Name"),
        description=_("Default singer name of voicebank"),
    )



from enum import Enum


class ActuateValue(Enum):
    ON_REQUEST = "onRequest"
    ON_LOAD = "onLoad"
    OTHER = "other"
    NONE = "none"


class ShowValue(Enum):
    NEW = "new"
    REPLACE = "replace"
    EMBED = "embed"
    OTHER = "other"
    NONE = "none"


class TypeValue(Enum):
    SIMPLE = "simple"


class LangValue(Enum):
    VALUE = ""


class SpaceValue(Enum):
    DEFAULT = "default"
    PRESERVE = "preserve"


class AboveBelow(Enum):
    """
    The above-below type is used to indicate whether one element appears above or
    below another element.
    """

    ABOVE = "above"
    BELOW = "below"


class AccidentalValue(Enum):
    """The accidental-value type represents notated accidentals supported by
    MusicXML.

    In the MusicXML 2.0 DTD this was a string with values that could be
    included. The XSD strengthens the data typing to an enumerated list.
    The quarter- and three-quarters- accidentals are Tartini-style
    quarter-tone accidentals. The -down and -up accidentals are quarter-
    tone accidentals that include arrows pointing down or up. The slash-
    accidentals are used in Turkish classical music. The numbered sharp
    and flat accidentals are superscripted versions of the accidental
    signs, used in Turkish folk music. The sori and koron accidentals
    are microtonal sharp and flat accidentals used in Iranian and
    Persian music. The other accidental covers accidentals other than
    those listed here. It is usually used in combination with the smufl
    attribute to specify a particular SMuFL accidental. The smufl
    attribute may be used with any accidental value to help specify the
    appearance of symbols that share the same MusicXML semantics.
    """

    SHARP = "sharp"
    NATURAL = "natural"
    FLAT = "flat"
    DOUBLE_SHARP = "double-sharp"
    SHARP_SHARP = "sharp-sharp"
    FLAT_FLAT = "flat-flat"
    NATURAL_SHARP = "natural-sharp"
    NATURAL_FLAT = "natural-flat"
    QUARTER_FLAT = "quarter-flat"
    QUARTER_SHARP = "quarter-sharp"
    THREE_QUARTERS_FLAT = "three-quarters-flat"
    THREE_QUARTERS_SHARP = "three-quarters-sharp"
    SHARP_DOWN = "sharp-down"
    SHARP_UP = "sharp-up"
    NATURAL_DOWN = "natural-down"
    NATURAL_UP = "natural-up"
    FLAT_DOWN = "flat-down"
    FLAT_UP = "flat-up"
    DOUBLE_SHARP_DOWN = "double-sharp-down"
    DOUBLE_SHARP_UP = "double-sharp-up"
    FLAT_FLAT_DOWN = "flat-flat-down"
    FLAT_FLAT_UP = "flat-flat-up"
    ARROW_DOWN = "arrow-down"
    ARROW_UP = "arrow-up"
    TRIPLE_SHARP = "triple-sharp"
    TRIPLE_FLAT = "triple-flat"
    SLASH_QUARTER_SHARP = "slash-quarter-sharp"
    SLASH_SHARP = "slash-sharp"
    SLASH_FLAT = "slash-flat"
    DOUBLE_SLASH_FLAT = "double-slash-flat"
    SHARP_1 = "sharp-1"
    SHARP_2 = "sharp-2"
    SHARP_3 = "sharp-3"
    SHARP_5 = "sharp-5"
    FLAT_1 = "flat-1"
    FLAT_2 = "flat-2"
    FLAT_3 = "flat-3"
    FLAT_4 = "flat-4"
    SORI = "sori"
    KORON = "koron"
    OTHER = "other"


class ArrowDirection(Enum):
    """
    The arrow-direction type represents the direction in which an arrow points,
    using Unicode arrow terminology.
    """

    LEFT = "left"
    UP = "up"
    RIGHT = "right"
    DOWN = "down"
    NORTHWEST = "northwest"
    NORTHEAST = "northeast"
    SOUTHEAST = "southeast"
    SOUTHWEST = "southwest"
    LEFT_RIGHT = "left right"
    UP_DOWN = "up down"
    NORTHWEST_SOUTHEAST = "northwest southeast"
    NORTHEAST_SOUTHWEST = "northeast southwest"
    OTHER = "other"


class ArrowStyle(Enum):
    """The arrow-style type represents the style of an arrow, using Unicode arrow
    terminology.

    Filled and hollow arrows indicate polygonal single arrows. Paired
    arrows are duplicate single arrows in the same direction. Combined
    arrows apply to double direction arrows like left right, indicating
    that an arrow in one direction should be combined with an arrow in
    the other direction.
    """

    SINGLE = "single"
    DOUBLE = "double"
    FILLED = "filled"
    HOLLOW = "hollow"
    PAIRED = "paired"
    COMBINED = "combined"
    OTHER = "other"


class BackwardForward(Enum):
    """The backward-forward type is used to specify repeat directions.

    The start of the repeat has a forward direction while the end of the
    repeat has a backward direction.
    """

    BACKWARD = "backward"
    FORWARD = "forward"


class BarStyle(Enum):
    """The bar-style type represents barline style information.

    Choices are regular, dotted, dashed, heavy, light-light, light-
    heavy, heavy-light, heavy-heavy, tick (a short stroke through the
    top line), short (a partial barline between the 2nd and 4th lines),
    and none.
    """

    REGULAR = "regular"
    DOTTED = "dotted"
    DASHED = "dashed"
    HEAVY = "heavy"
    LIGHT_LIGHT = "light-light"
    LIGHT_HEAVY = "light-heavy"
    HEAVY_LIGHT = "heavy-light"
    HEAVY_HEAVY = "heavy-heavy"
    TICK = "tick"
    SHORT = "short"
    NONE = "none"


class BeamValue(Enum):
    """
    The beam-value type represents the type of beam associated with each of 8 beam
    levels (up to 1024th notes) available for each note.
    """

    BEGIN = "begin"
    CONTINUE = "continue"
    END = "end"
    FORWARD_HOOK = "forward hook"
    BACKWARD_HOOK = "backward hook"


class BeaterValue(Enum):
    """The beater-value type represents pictograms for beaters, mallets, and sticks
    that do not have different materials represented in the pictogram.

    The finger and hammer values are in addition to Stone's list.
    """

    BOW = "bow"
    CHIME_HAMMER = "chime hammer"
    COIN = "coin"
    DRUM_STICK = "drum stick"
    FINGER = "finger"
    FINGERNAIL = "fingernail"
    FIST = "fist"
    GUIRO_SCRAPER = "guiro scraper"
    HAMMER = "hammer"
    HAND = "hand"
    JAZZ_STICK = "jazz stick"
    KNITTING_NEEDLE = "knitting needle"
    METAL_HAMMER = "metal hammer"
    SLIDE_BRUSH_ON_GONG = "slide brush on gong"
    SNARE_STICK = "snare stick"
    SPOON_MALLET = "spoon mallet"
    SUPERBALL = "superball"
    TRIANGLE_BEATER = "triangle beater"
    TRIANGLE_BEATER_PLAIN = "triangle beater plain"
    WIRE_BRUSH = "wire brush"


class BendShape(Enum):
    """
    The bend-shape type distinguishes between the angled bend symbols commonly used
    in standard notation and the curved bend symbols commonly used in both
    tablature and standard notation.
    """

    ANGLED = "angled"
    CURVED = "curved"


class BreathMarkValue(Enum):
    """
    The breath-mark-value type represents the symbol used for a breath mark.
    """

    VALUE = ""
    COMMA = "comma"
    TICK = "tick"
    UPBOW = "upbow"
    SALZEDO = "salzedo"


class CaesuraValue(Enum):
    """
    The caesura-value type represents the shape of the caesura sign.
    """

    NORMAL = "normal"
    THICK = "thick"
    SHORT = "short"
    CURVED = "curved"
    SINGLE = "single"
    VALUE = ""


class CancelLocation(Enum):
    """The cancel-location type is used to indicate where a key signature cancellation appears relative to a new key signature: to the left, to the right, or before the barline and to the left. It is left by default. For mid-measure key elements, a cancel-location of before-barline should be treated like a cancel-location of left."""

    LEFT = "left"
    RIGHT = "right"
    BEFORE_BARLINE = "before-barline"


class CircularArrow(Enum):
    """
    The circular-arrow type represents the direction in which a circular arrow
    points, using Unicode arrow terminology.
    """

    CLOCKWISE = "clockwise"
    ANTICLOCKWISE = "anticlockwise"


class ClefSign(Enum):
    """The clef-sign type represents the different clef symbols.

    The jianpu sign indicates that the music that follows should be in
    jianpu numbered notation, just as the TAB sign indicates that the
    music that follows should be in tablature notation. Unlike TAB, a
    jianpu sign does not correspond to a visual clef notation. The none
    sign is deprecated as of MusicXML 4.0. Use the clef element's print-
    object attribute instead. When the none sign is used, notes should
    be displayed as if in treble clef.
    """

    G = "G"
    F = "F"
    C = "C"
    PERCUSSION = "percussion"
    TAB = "TAB"
    JIANPU = "jianpu"
    NONE = "none"


class CssFontSize(Enum):
    """
    The css-font-size type includes the CSS font sizes used as an alternative to a
    numeric point size.
    """

    XX_SMALL = "xx-small"
    X_SMALL = "x-small"
    SMALL = "small"
    MEDIUM = "medium"
    LARGE = "large"
    X_LARGE = "x-large"
    XX_LARGE = "xx-large"


class DegreeSymbolValue(Enum):
    """
    The degree-symbol-value type indicates which symbol should be used in
    specifying a degree.
    """

    MAJOR = "major"
    MINOR = "minor"
    AUGMENTED = "augmented"
    DIMINISHED = "diminished"
    HALF_DIMINISHED = "half-diminished"


class DegreeTypeValue(Enum):
    """
    The degree-type-value type indicates whether the current degree element is an
    addition, alteration, or subtraction to the kind of the current chord in the
    harmony element.
    """

    ADD = "add"
    ALTER = "alter"
    SUBTRACT = "subtract"


class EffectValue(Enum):
    """The effect-value type represents pictograms for sound effect percussion
    instruments.

    The cannon, lotus flute, and megaphone values are in addition to
    Stone's list.
    """

    ANVIL = "anvil"
    AUTO_HORN = "auto horn"
    BIRD_WHISTLE = "bird whistle"
    CANNON = "cannon"
    DUCK_CALL = "duck call"
    GUN_SHOT = "gun shot"
    KLAXON_HORN = "klaxon horn"
    LIONS_ROAR = "lions roar"
    LOTUS_FLUTE = "lotus flute"
    MEGAPHONE = "megaphone"
    POLICE_WHISTLE = "police whistle"
    SIREN = "siren"
    SLIDE_WHISTLE = "slide whistle"
    THUNDER_SHEET = "thunder sheet"
    WIND_MACHINE = "wind machine"
    WIND_WHISTLE = "wind whistle"


class EnclosureShape(Enum):
    """The enclosure-shape type describes the shape and presence / absence of an
    enclosure around text or symbols.

    A bracket enclosure is similar to a rectangle with the bottom line
    missing, as is common in jazz notation. An inverted-bracket
    enclosure is similar to a rectangle with the top line missing.
    """

    RECTANGLE = "rectangle"
    SQUARE = "square"
    OVAL = "oval"
    CIRCLE = "circle"
    BRACKET = "bracket"
    INVERTED_BRACKET = "inverted-bracket"
    TRIANGLE = "triangle"
    DIAMOND = "diamond"
    PENTAGON = "pentagon"
    HEXAGON = "hexagon"
    HEPTAGON = "heptagon"
    OCTAGON = "octagon"
    NONAGON = "nonagon"
    DECAGON = "decagon"
    NONE = "none"


class Fan(Enum):
    """
    The fan type represents the type of beam fanning present on a note, used to
    represent accelerandos and ritardandos.
    """

    ACCEL = "accel"
    RIT = "rit"
    NONE = "none"


class FermataShape(Enum):
    """The fermata-shape type represents the shape of the fermata sign.

    The empty value is equivalent to the normal value.
    """

    NORMAL = "normal"
    ANGLED = "angled"
    SQUARE = "square"
    DOUBLE_ANGLED = "double-angled"
    DOUBLE_SQUARE = "double-square"
    DOUBLE_DOT = "double-dot"
    HALF_CURVE = "half-curve"
    CURLEW = "curlew"
    VALUE = ""


class FontStyle(Enum):
    """
    The font-style type represents a simplified version of the CSS font-style
    property.
    """

    NORMAL = "normal"
    ITALIC = "italic"


class FontWeight(Enum):
    """
    The font-weight type represents a simplified version of the CSS font-weight
    property.
    """

    NORMAL = "normal"
    BOLD = "bold"


class GlassValue(Enum):
    """
    The glass-value type represents pictograms for glass percussion instruments.
    """

    GLASS_HARMONICA = "glass harmonica"
    GLASS_HARP = "glass harp"
    WIND_CHIMES = "wind chimes"


class GroupBarlineValue(Enum):
    """
    The group-barline-value type indicates if the group should have common
    barlines.
    """

    YES = "yes"
    NO = "no"
    MENSURSTRICH = "Mensurstrich"


class GroupSymbolValue(Enum):
    """
    The group-symbol-value type indicates how the symbol for a group or multi-staff
    part is indicated in the score.
    """

    NONE = "none"
    BRACE = "brace"
    LINE = "line"
    BRACKET = "bracket"
    SQUARE = "square"


class HandbellValue(Enum):
    """
    The handbell-value type represents the type of handbell technique being
    notated.
    """

    BELLTREE = "belltree"
    DAMP = "damp"
    ECHO = "echo"
    GYRO = "gyro"
    HAND_MARTELLATO = "hand martellato"
    MALLET_LIFT = "mallet lift"
    MALLET_TABLE = "mallet table"
    MARTELLATO = "martellato"
    MARTELLATO_LIFT = "martellato lift"
    MUTED_MARTELLATO = "muted martellato"
    PLUCK_LIFT = "pluck lift"
    SWING = "swing"


class HarmonClosedLocation(Enum):
    """
    The harmon-closed-location type indicates which portion of the symbol is filled
    in when the corresponding harmon-closed-value is half.
    """

    RIGHT = "right"
    BOTTOM = "bottom"
    LEFT = "left"
    TOP = "top"


class HarmonClosedValue(Enum):
    """
    The harmon-closed-value type represents whether the harmon mute is closed,
    open, or half-open.
    """

    YES = "yes"
    NO = "no"
    HALF = "half"


class HarmonyArrangement(Enum):
    """The harmony-arrangement type indicates how stacked chords and bass notes are
    displayed within a harmony element.

    The vertical value specifies that the second element appears below
    the first. The horizontal value specifies that the second element
    appears to the right of the first. The diagonal value specifies that
    the second element appears both below and to the right of the first.
    """

    VERTICAL = "vertical"
    HORIZONTAL = "horizontal"
    DIAGONAL = "diagonal"


class HarmonyType(Enum):
    """The harmony-type type differentiates different types of harmonies when
    alternate harmonies are possible.

    Explicit harmonies have all note present in the music; implied have
    some notes missing but implied; alternate represents alternate
    analyses.
    """

    EXPLICIT = "explicit"
    IMPLIED = "implied"
    ALTERNATE = "alternate"


class HoleClosedLocation(Enum):
    """
    The hole-closed-location type indicates which portion of the hole is filled in
    when the corresponding hole-closed-value is half.
    """

    RIGHT = "right"
    BOTTOM = "bottom"
    LEFT = "left"
    TOP = "top"


class HoleClosedValue(Enum):
    """
    The hole-closed-value type represents whether the hole is closed, open, or
    half-open.
    """

    YES = "yes"
    NO = "no"
    HALF = "half"


class KindValue(Enum):
    """A kind-value indicates the type of chord.

    Degree elements can then add, subtract, or alter from these starting points. Values include:
    Triads:
    major (major third, perfect fifth)
    minor (minor third, perfect fifth)
    augmented (major third, augmented fifth)
    diminished (minor third, diminished fifth)
    Sevenths:
    dominant (major triad, minor seventh)
    major-seventh (major triad, major seventh)
    minor-seventh (minor triad, minor seventh)
    diminished-seventh (diminished triad, diminished seventh)
    augmented-seventh (augmented triad, minor seventh)
    half-diminished (diminished triad, minor seventh)
    major-minor (minor triad, major seventh)
    Sixths:
    major-sixth (major triad, added sixth)
    minor-sixth (minor triad, added sixth)
    Ninths:
    dominant-ninth (dominant-seventh, major ninth)
    major-ninth (major-seventh, major ninth)
    minor-ninth (minor-seventh, major ninth)
    11ths (usually as the basis for alteration):
    dominant-11th (dominant-ninth, perfect 11th)
    major-11th (major-ninth, perfect 11th)
    minor-11th (minor-ninth, perfect 11th)
    13ths (usually as the basis for alteration):
    dominant-13th (dominant-11th, major 13th)
    major-13th (major-11th, major 13th)
    minor-13th (minor-11th, major 13th)
    Suspended:
    suspended-second (major second, perfect fifth)
    suspended-fourth (perfect fourth, perfect fifth)
    Functional sixths:
    Neapolitan
    Italian
    French
    German
    Other:
    pedal (pedal-point bass)
    power (perfect fifth)
    Tristan
    The "other" kind is used when the harmony is entirely composed of add elements.
    The "none" kind is used to explicitly encode absence of chords or functional harmony. In this case, the root, numeral, or function element has no meaning. When using the root or numeral element, the root-step or numeral-step text attribute should be set to the empty string to keep the root or numeral from being displayed.
    """

    MAJOR = "major"
    MINOR = "minor"
    AUGMENTED = "augmented"
    DIMINISHED = "diminished"
    DOMINANT = "dominant"
    MAJOR_SEVENTH = "major-seventh"
    MINOR_SEVENTH = "minor-seventh"
    DIMINISHED_SEVENTH = "diminished-seventh"
    AUGMENTED_SEVENTH = "augmented-seventh"
    HALF_DIMINISHED = "half-diminished"
    MAJOR_MINOR = "major-minor"
    MAJOR_SIXTH = "major-sixth"
    MINOR_SIXTH = "minor-sixth"
    DOMINANT_NINTH = "dominant-ninth"
    MAJOR_NINTH = "major-ninth"
    MINOR_NINTH = "minor-ninth"
    DOMINANT_11TH = "dominant-11th"
    MAJOR_11TH = "major-11th"
    MINOR_11TH = "minor-11th"
    DOMINANT_13TH = "dominant-13th"
    MAJOR_13TH = "major-13th"
    MINOR_13TH = "minor-13th"
    SUSPENDED_SECOND = "suspended-second"
    SUSPENDED_FOURTH = "suspended-fourth"
    NEAPOLITAN = "Neapolitan"
    ITALIAN = "Italian"
    FRENCH = "French"
    GERMAN = "German"
    PEDAL = "pedal"
    POWER = "power"
    TRISTAN = "Tristan"
    OTHER = "other"
    NONE = "none"


class LeftCenterRight(Enum):
    """
    The left-center-right type is used to define horizontal alignment and text
    justification.
    """

    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class LeftRight(Enum):
    """
    The left-right type is used to indicate whether one element appears to the left
    or the right of another element.
    """

    LEFT = "left"
    RIGHT = "right"


class LineEnd(Enum):
    """
    The line-end type specifies if there is a jog up or down (or both), an arrow,
    or nothing at the start or end of a bracket.
    """

    UP = "up"
    DOWN = "down"
    BOTH = "both"
    ARROW = "arrow"
    NONE = "none"


class LineLength(Enum):
    """
    The line-length type distinguishes between different line lengths for doit,
    falloff, plop, and scoop articulations.
    """

    SHORT = "short"
    MEDIUM = "medium"
    LONG = "long"


class LineShape(Enum):
    """
    The line-shape type distinguishes between straight and curved lines.
    """

    STRAIGHT = "straight"
    CURVED = "curved"


class LineType(Enum):
    """
    The line-type type distinguishes between solid, dashed, dotted, and wavy lines.
    """

    SOLID = "solid"
    DASHED = "dashed"
    DOTTED = "dotted"
    WAVY = "wavy"


class MarginType(Enum):
    """
    The margin-type type specifies whether margins apply to even page, odd pages,
    or both.
    """

    ODD = "odd"
    EVEN = "even"
    BOTH = "both"


class MeasureNumberingValue(Enum):
    """The measure-numbering-value type describes how measure numbers are displayed on this part: no numbers, numbers every measure, or numbers every system."""

    NONE = "none"
    MEASURE = "measure"
    SYSTEM = "system"


class MembraneValue(Enum):
    """
    The membrane-value type represents pictograms for membrane percussion
    instruments.
    """

    BASS_DRUM = "bass drum"
    BASS_DRUM_ON_SIDE = "bass drum on side"
    BONGOS = "bongos"
    CHINESE_TOMTOM = "Chinese tomtom"
    CONGA_DRUM = "conga drum"
    CUICA = "cuica"
    GOBLET_DRUM = "goblet drum"
    INDO_AMERICAN_TOMTOM = "Indo-American tomtom"
    JAPANESE_TOMTOM = "Japanese tomtom"
    MILITARY_DRUM = "military drum"
    SNARE_DRUM = "snare drum"
    SNARE_DRUM_SNARES_OFF = "snare drum snares off"
    TABLA = "tabla"
    TAMBOURINE = "tambourine"
    TENOR_DRUM = "tenor drum"
    TIMBALES = "timbales"
    TOMTOM = "tomtom"


class MetalValue(Enum):
    """The metal-value type represents pictograms for metal percussion instruments.

    The hi-hat value refers to a pictogram like Stone's high-hat cymbals
    but without the long vertical line at the bottom.
    """

    AGOGO = "agogo"
    ALMGLOCKEN = "almglocken"
    BELL = "bell"
    BELL_PLATE = "bell plate"
    BELL_TREE = "bell tree"
    BRAKE_DRUM = "brake drum"
    CENCERRO = "cencerro"
    CHAIN_RATTLE = "chain rattle"
    CHINESE_CYMBAL = "Chinese cymbal"
    COWBELL = "cowbell"
    CRASH_CYMBALS = "crash cymbals"
    CROTALE = "crotale"
    CYMBAL_TONGS = "cymbal tongs"
    DOMED_GONG = "domed gong"
    FINGER_CYMBALS = "finger cymbals"
    FLEXATONE = "flexatone"
    GONG = "gong"
    HI_HAT = "hi-hat"
    HIGH_HAT_CYMBALS = "high-hat cymbals"
    HANDBELL = "handbell"
    JAW_HARP = "jaw harp"
    JINGLE_BELLS = "jingle bells"
    MUSICAL_SAW = "musical saw"
    SHELL_BELLS = "shell bells"
    SISTRUM = "sistrum"
    SIZZLE_CYMBAL = "sizzle cymbal"
    SLEIGH_BELLS = "sleigh bells"
    SUSPENDED_CYMBAL = "suspended cymbal"
    TAM_TAM = "tam tam"
    TAM_TAM_WITH_BEATER = "tam tam with beater"
    TRIANGLE = "triangle"
    VIETNAMESE_HAT = "Vietnamese hat"


class Mute(Enum):
    """The mute type represents muting for different instruments, including brass,
    winds, and strings.

    The on and off values are used for undifferentiated mutes. The
    remaining values represent specific mutes.
    """

    ON = "on"
    OFF = "off"
    STRAIGHT = "straight"
    CUP = "cup"
    HARMON_NO_STEM = "harmon-no-stem"
    HARMON_STEM = "harmon-stem"
    BUCKET = "bucket"
    PLUNGER = "plunger"
    HAT = "hat"
    SOLOTONE = "solotone"
    PRACTICE = "practice"
    STOP_MUTE = "stop-mute"
    STOP_HAND = "stop-hand"
    ECHO = "echo"
    PALM = "palm"


class NoteSizeType(Enum):
    """The note-size-type type indicates the type of note being defined by a note-
    size element.

    The grace-cue type is used for notes of grace-cue size. The grace
    type is used for notes of cue size that include a grace element. The
    cue type is used for all other notes with cue size, whether defined
    explicitly or implicitly via a cue element. The large type is used
    for notes of large size.
    """

    CUE = "cue"
    GRACE = "grace"
    GRACE_CUE = "grace-cue"
    LARGE = "large"


class NoteTypeValue(Enum):
    """
    The note-type-value type is used for the MusicXML type element and represents
    the graphic note type, from 1024th (shortest) to maxima (longest).
    """

    VALUE_1024TH = "1024th"
    VALUE_512TH = "512th"
    VALUE_256TH = "256th"
    VALUE_128TH = "128th"
    VALUE_64TH = "64th"
    VALUE_32ND = "32nd"
    VALUE_16TH = "16th"
    EIGHTH = "eighth"
    QUARTER = "quarter"
    HALF = "half"
    WHOLE = "whole"
    BREVE = "breve"
    LONG = "long"
    MAXIMA = "maxima"


class NoteheadValue(Enum):
    """The notehead-value type indicates shapes other than the open and closed
    ovals associated with note durations.

    The values do, re, mi, fa, fa up, so, la, and ti correspond to
    Aikin's 7-shape system.  The fa up shape is typically used with
    upstems; the fa shape is typically used with downstems or no stems.
    The arrow shapes differ from triangle and inverted triangle by being
    centered on the stem. Slashed and back slashed notes include both
    the normal notehead and a slash. The triangle shape has the tip of
    the triangle pointing up; the inverted triangle shape has the tip of
    the triangle pointing down. The left triangle shape is a right
    triangle with the hypotenuse facing up and to the left. The other
    notehead covers noteheads other than those listed here. It is
    usually used in combination with the smufl attribute to specify a
    particular SMuFL notehead. The smufl attribute may be used with any
    notehead value to help specify the appearance of symbols that share
    the same MusicXML semantics. Noteheads in the SMuFL Note name
    noteheads and Note name noteheads supplement ranges (U+E150-U+E1AF
    and U+EEE0-U+EEFF) should not use the smufl attribute or the "other"
    value, but instead use the notehead-text element.
    """

    SLASH = "slash"
    TRIANGLE = "triangle"
    DIAMOND = "diamond"
    SQUARE = "square"
    CROSS = "cross"
    X = "x"
    CIRCLE_X = "circle-x"
    INVERTED_TRIANGLE = "inverted triangle"
    ARROW_DOWN = "arrow down"
    ARROW_UP = "arrow up"
    CIRCLED = "circled"
    SLASHED = "slashed"
    BACK_SLASHED = "back slashed"
    NORMAL = "normal"
    CLUSTER = "cluster"
    CIRCLE_DOT = "circle dot"
    LEFT_TRIANGLE = "left triangle"
    RECTANGLE = "rectangle"
    NONE = "none"
    DO = "do"
    RE = "re"
    MI = "mi"
    FA = "fa"
    FA_UP = "fa up"
    SO = "so"
    LA = "la"
    TI = "ti"
    OTHER = "other"


class NumberOrNormalValue(Enum):
    NORMAL = "normal"


class NumeralMode(Enum):
    """The numeral-mode type specifies the mode similar to the mode type, but with
    a restricted set of values.

    The different minor values are used to interpret numeral-root values
    of 6 and 7 when present in a minor key. The harmonic minor value
    sharpens the 7 and the melodic minor value sharpens both 6 and 7. If
    a minor mode is used without qualification, either in the mode or
    numeral-mode elements, natural minor is used.
    """

    MAJOR = "major"
    MINOR = "minor"
    NATURAL_MINOR = "natural minor"
    MELODIC_MINOR = "melodic minor"
    HARMONIC_MINOR = "harmonic minor"


class OnOff(Enum):
    """
    The on-off type is used for notation elements such as string mutes.
    """

    ON = "on"
    OFF = "off"


class OverUnder(Enum):
    """
    The over-under type is used to indicate whether the tips of curved lines such
    as slurs and ties are overhand (tips down) or underhand (tips up).
    """

    OVER = "over"
    UNDER = "under"


class PedalType(Enum):
    """The pedal-type simple type is used to distinguish types of pedal directions.

    The start value indicates the start of a damper pedal, while the
    sostenuto value indicates the start of a sostenuto pedal. The other
    values can be used with either the damper or sostenuto pedal. The
    soft pedal is not included here because there is no special symbol
    or graphic used for it beyond what can be specified with words and
    bracket elements. The change, continue, discontinue, and resume
    types are used when the line attribute is yes. The change type
    indicates a pedal lift and retake indicated with an inverted V
    marking. The continue type allows more precise formatting across
    system breaks and for more complex pedaling lines. The discontinue
    type indicates the end of a pedal line that does not include the
    explicit lift represented by the stop type. The resume type
    indicates the start of a pedal line that does not include the
    downstroke represented by the start type. It can be used when a line
    resumes after being discontinued, or to start a pedal line that is
    preceded by a text or symbol representation of the pedal.
    """

    START = "start"
    STOP = "stop"
    SOSTENUTO = "sostenuto"
    CHANGE = "change"
    CONTINUE = "continue"
    DISCONTINUE = "discontinue"
    RESUME = "resume"


class PitchedValue(Enum):
    """The pitched-value type represents pictograms for pitched percussion
    instruments.

    The chimes and tubular chimes values distinguish the single-line and
    double-line versions of the pictogram.
    """

    CELESTA = "celesta"
    CHIMES = "chimes"
    GLOCKENSPIEL = "glockenspiel"
    LITHOPHONE = "lithophone"
    MALLET = "mallet"
    MARIMBA = "marimba"
    STEEL_DRUMS = "steel drums"
    TUBAPHONE = "tubaphone"
    TUBULAR_CHIMES = "tubular chimes"
    VIBRAPHONE = "vibraphone"
    XYLOPHONE = "xylophone"


class PositiveIntegerOrEmptyValue(Enum):
    VALUE = ""


class PrincipalVoiceSymbol(Enum):
    """The principal-voice-symbol type represents the type of symbol used to
    indicate a principal or secondary voice.

    The "plain" value represents a plain square bracket. The value of
    "none" is used for analysis markup when the principal-voice element
    does not have a corresponding appearance in the score.
    """

    HAUPTSTIMME = "Hauptstimme"
    NEBENSTIMME = "Nebenstimme"
    PLAIN = "plain"
    NONE = "none"


class RightLeftMiddle(Enum):
    """
    The right-left-middle type is used to specify barline location.
    """

    RIGHT = "right"
    LEFT = "left"
    MIDDLE = "middle"


class SemiPitched(Enum):
    """
    The semi-pitched type represents categories of indefinite pitch for percussion
    instruments.
    """

    HIGH = "high"
    MEDIUM_HIGH = "medium-high"
    MEDIUM = "medium"
    MEDIUM_LOW = "medium-low"
    LOW = "low"
    VERY_LOW = "very-low"


class ShowFrets(Enum):
    """The show-frets type indicates whether to show tablature frets as numbers (0,
    1, 2) or letters (a, b, c).

    The default choice is numbers.
    """

    NUMBERS = "numbers"
    LETTERS = "letters"


class ShowTuplet(Enum):
    """
    The show-tuplet type indicates whether to show a part of a tuplet relating to
    the tuplet-actual element, both the tuplet-actual and tuplet-normal elements,
    or neither.
    """

    ACTUAL = "actual"
    BOTH = "both"
    NONE = "none"


class StaffDivideSymbol(Enum):
    """The staff-divide-symbol type is used for staff division symbols.

    The down, up, and up-down values correspond to SMuFL code points
    U+E00B, U+E00C, and U+E00D respectively.
    """

    DOWN = "down"
    UP = "up"
    UP_DOWN = "up-down"


class StaffType(Enum):
    """The staff-type value can be ossia, editorial, cue, alternate, or regular.

    An ossia staff represents music that can be played instead of what
    appears on the regular staff. An editorial staff also represents
    musical alternatives, but is created by an editor rather than the
    composer. It can be used for suggested interpretations or
    alternatives from other sources. A cue staff represents music from
    another part. An alternate staff shares the same music as the prior
    staff, but displayed differently (e.g., treble and bass clef,
    standard notation and tablature). It is not included in playback. An
    alternate staff provides more information to an application reading
    a file than encoding the same music in separate parts, so its use is
    preferred in this situation if feasible. A regular staff is the
    standard default staff-type.
    """

    OSSIA = "ossia"
    EDITORIAL = "editorial"
    CUE = "cue"
    ALTERNATE = "alternate"
    REGULAR = "regular"


class StartNote(Enum):
    """
    The start-note type describes the starting note of trills and mordents for
    playback, relative to the current note.
    """

    UPPER = "upper"
    MAIN = "main"
    BELOW = "below"


class StartStop(Enum):
    """The start-stop type is used for an attribute of musical elements that can
    either start or stop, such as tuplets.

    The values of start and stop refer to how an element appears in
    musical score order, not in MusicXML document order. An element with
    a stop attribute may precede the corresponding element with a start
    attribute within a MusicXML document. This is particularly common in
    multi-staff music. For example, the stopping point for a tuplet may
    appear in staff 1 before the starting point for the tuplet appears
    in staff 2 later in the document. When multiple elements with the
    same tag are used within the same note, their order within the
    MusicXML document should match the musical score order.
    """

    START = "start"
    STOP = "stop"


class StartStopContinue(Enum):
    """The start-stop-continue type is used for an attribute of musical elements
    that can either start or stop, but also need to refer to an intermediate point
    in the symbol, as for complex slurs or for formatting of symbols across system
    breaks.

    The values of start, stop, and continue refer to how an element
    appears in musical score order, not in MusicXML document order. An
    element with a stop attribute may precede the corresponding element
    with a start attribute within a MusicXML document. This is
    particularly common in multi-staff music. For example, the stopping
    point for a slur may appear in staff 1 before the starting point for
    the slur appears in staff 2 later in the document. When multiple
    elements with the same tag are used within the same note, their
    order within the MusicXML document should match the musical score
    order. For example, a note that marks both the end of one slur and
    the start of a new slur should have the incoming slur element with a
    type of stop precede the outgoing slur element with a type of start.
    """

    START = "start"
    STOP = "stop"
    CONTINUE = "continue"


class StartStopDiscontinue(Enum):
    """The start-stop-discontinue type is used to specify ending types.

    Typically, the start type is associated with the left barline of the
    first measure in an ending. The stop and discontinue types are
    associated with the right barline of the last measure in an ending.
    Stop is used when the ending mark concludes with a downward jog, as
    is typical for first endings. Discontinue is used when there is no
    downward jog, as is typical for second endings that do not conclude
    a piece.
    """

    START = "start"
    STOP = "stop"
    DISCONTINUE = "discontinue"


class StartStopSingle(Enum):
    """The start-stop-single type is used for an attribute of musical elements that
    can be used for either multi-note or single-note musical elements, as for
    groupings.

    When multiple elements with the same tag are used within the same
    note, their order within the MusicXML document should match the
    musical score order.
    """

    START = "start"
    STOP = "stop"
    SINGLE = "single"


class StemValue(Enum):
    """
    The stem-value type represents the notated stem direction.
    """

    DOWN = "down"
    UP = "up"
    DOUBLE = "double"
    NONE = "none"


class Step(Enum):
    """
    The step type represents a step of the diatonic scale, represented using the
    English letters A through G.
    """

    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"


class StickLocation(Enum):
    """
    The stick-location type represents pictograms for the location of sticks,
    beaters, or mallets on cymbals, gongs, drums, and other instruments.
    """

    CENTER = "center"
    RIM = "rim"
    CYMBAL_BELL = "cymbal bell"
    CYMBAL_EDGE = "cymbal edge"


class StickMaterial(Enum):
    """
    The stick-material type represents the material being displayed in a stick
    pictogram.
    """

    SOFT = "soft"
    MEDIUM = "medium"
    HARD = "hard"
    SHADED = "shaded"
    X = "x"


class StickType(Enum):
    """
    The stick-type type represents the shape of pictograms where the material in
    the stick, mallet, or beater is represented in the pictogram.
    """

    BASS_DRUM = "bass drum"
    DOUBLE_BASS_DRUM = "double bass drum"
    GLOCKENSPIEL = "glockenspiel"
    GUM = "gum"
    HAMMER = "hammer"
    SUPERBALL = "superball"
    TIMPANI = "timpani"
    WOUND = "wound"
    XYLOPHONE = "xylophone"
    YARN = "yarn"


class SwingTypeValue(Enum):
    """
    The swing-type-value type specifies the note type, either eighth or 16th, to
    which the ratio defined in the swing element is applied.
    """

    VALUE_16TH = "16th"
    EIGHTH = "eighth"


class Syllabic(Enum):
    """Lyric hyphenation is indicated by the syllabic type.

    The single, begin, end, and middle values represent single-syllable
    words, word-beginning syllables, word-ending syllables, and mid-word
    syllables, respectively.
    """

    SINGLE = "single"
    BEGIN = "begin"
    END = "end"
    MIDDLE = "middle"


class SymbolSize(Enum):
    """
    The symbol-size type is used to distinguish between full, cue sized, grace cue
    sized, and oversized symbols.
    """

    FULL = "full"
    CUE = "cue"
    GRACE_CUE = "grace-cue"
    LARGE = "large"


class SyncType(Enum):
    """The sync-type type specifies the style that a score following application
    should use to synchronize an accompaniment with a performer.

    The none type indicates no synchronization to the performer. The
    tempo type indicates synchronization based on the performer tempo
    rather than individual events in the score. The event type indicates
    synchronization by following the performance of individual events in
    the score rather than the performer tempo. The mostly-tempo and
    mostly-event types combine these two approaches, with mostly-tempo
    giving more weight to tempo and mostly-event giving more weight to
    performed events. The always-event type provides the strictest
    synchronization by not being forgiving of missing performed events.
    """

    NONE = "none"
    TEMPO = "tempo"
    MOSTLY_TEMPO = "mostly-tempo"
    MOSTLY_EVENT = "mostly-event"
    EVENT = "event"
    ALWAYS_EVENT = "always-event"


class SystemRelation(Enum):
    """The system-relation type distinguishes elements that are associated with a
    system rather than the particular part where the element appears.

    A value of only-top indicates that the element should appear only on
    the top part of the current system. A value of also-top indicates
    that the element should appear on both the current part and the top
    part of the current system. If this value appears in a score, when
    parts are created the element should only appear once in this part,
    not twice. A value of none indicates that the element is associated
    only with the current part, not with the system.
    """

    ONLY_TOP = "only-top"
    ALSO_TOP = "also-top"
    NONE = "none"


class SystemRelationNumber(Enum):
    """The system-relation-number type distinguishes measure numbers that are
    associated with a system rather than the particular part where the element
    appears.

    A value of only-top or only-bottom indicates that the number should
    appear only on the top or bottom part of the current system,
    respectively. A value of also-top or also-bottom indicates that the
    number should appear on both the current part and the top or bottom
    part of the current system, respectively. If these values appear in
    a score, when parts are created the number should only appear once
    in this part, not twice. A value of none indicates that the number
    is associated only with the current part, not with the system.
    """

    ONLY_TOP = "only-top"
    ONLY_BOTTOM = "only-bottom"
    ALSO_TOP = "also-top"
    ALSO_BOTTOM = "also-bottom"
    NONE = "none"


class TapHand(Enum):
    """The tap-hand type represents the symbol to use for a tap element.

    The left and right values refer to the SMuFL guitarLeftHandTapping
    and guitarRightHandTapping glyphs respectively.
    """

    LEFT = "left"
    RIGHT = "right"


class TextDirection(Enum):
    """The text-direction type is used to adjust and override the Unicode
    bidirectional text algorithm, similar to the Directionality data category in
    the W3C Internationalization Tag Set recommendation.

    Values are ltr (left-to-right embed), rtl (right-to-left embed), lro
    (left-to-right bidi-override), and rlo (right-to-left bidi-
    override). The default value is ltr. This type is typically used by
    applications that store text in left-to-right visual order rather
    than logical order. Such applications can use the lro value to
    better communicate with other applications that more fully support
    bidirectional text.
    """

    LTR = "ltr"
    RTL = "rtl"
    LRO = "lro"
    RLO = "rlo"


class TiedType(Enum):
    """The tied-type type is used as an attribute of the tied element to specify
    where the visual representation of a tie begins and ends.

    A tied element which joins two notes of the same pitch can be
    specified with tied-type start on the first note and tied-type stop
    on the second note. To indicate a note should be undamped, use a
    single tied element with tied-type let-ring. For other ties that are
    visually attached to a single note, such as a tie leading into or
    out of a repeated section or coda, use two tied elements on the same
    note, one start and one stop. In start-stop cases, ties can add more
    elements using a continue type. This is typically used to specify
    the formatting of cross-system ties. When multiple elements with the
    same tag are used within the same note, their order within the
    MusicXML document should match the musical score order. For example,
    a note with a tie at the end of a first ending should have the tied
    element with a type of start precede the tied element with a type of
    stop.
    """

    START = "start"
    STOP = "stop"
    CONTINUE = "continue"
    LET_RING = "let-ring"


class TimeRelation(Enum):
    """
    The time-relation type indicates the symbol used to represent the
    interchangeable aspect of dual time signatures.
    """

    PARENTHESES = "parentheses"
    BRACKET = "bracket"
    EQUALS = "equals"
    SLASH = "slash"
    SPACE = "space"
    HYPHEN = "hyphen"


class TimeSeparator(Enum):
    """The time-separator type indicates how to display the arrangement between the
    beats and beat-type values in a time signature.

    The default value is none. The horizontal, diagonal, and vertical
    values represent horizontal, diagonal lower-left to upper-right, and
    vertical lines respectively. For these values, the beats and beat-
    type values are arranged on either side of the separator line. The
    none value represents no separator with the beats and beat-type
    arranged vertically. The adjacent value represents no separator with
    the beats and beat-type arranged horizontally.
    """

    NONE = "none"
    HORIZONTAL = "horizontal"
    DIAGONAL = "diagonal"
    VERTICAL = "vertical"
    ADJACENT = "adjacent"


class TimeSymbol(Enum):
    """The time-symbol type indicates how to display a time signature.

    The normal value is the usual fractional display, and is the implied
    symbol type if none is specified. Other options are the common and
    cut time symbols, as well as a single number with an implied
    denominator. The note symbol indicates that the beat-type should be
    represented with the corresponding downstem note rather than a
    number. The dotted-note symbol indicates that the beat-type should
    be represented with a dotted downstem note that corresponds to three
    times the beat-type value, and a numerator that is one third the
    beats value.
    """

    COMMON = "common"
    CUT = "cut"
    SINGLE_NUMBER = "single-number"
    NOTE = "note"
    DOTTED_NOTE = "dotted-note"
    NORMAL = "normal"


class TipDirection(Enum):
    """
    The tip-direction type represents the direction in which the tip of a stick or
    beater points, using Unicode arrow terminology.
    """

    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"
    NORTHWEST = "northwest"
    NORTHEAST = "northeast"
    SOUTHEAST = "southeast"
    SOUTHWEST = "southwest"


class TopBottom(Enum):
    """
    The top-bottom type is used to indicate the top or bottom part of a vertical
    shape like non-arpeggiate.
    """

    TOP = "top"
    BOTTOM = "bottom"


class TremoloType(Enum):
    """
    The tremolo-type is used to distinguish double-note, single-note, and
    unmeasured tremolos.
    """

    START = "start"
    STOP = "stop"
    SINGLE = "single"
    UNMEASURED = "unmeasured"


class TrillStep(Enum):
    """
    The trill-step type describes the alternating note of trills and mordents for
    playback, relative to the current note.
    """

    WHOLE = "whole"
    HALF = "half"
    UNISON = "unison"


class TwoNoteTurn(Enum):
    """
    The two-note-turn type describes the ending notes of trills and mordents for
    playback, relative to the current note.
    """

    WHOLE = "whole"
    HALF = "half"
    NONE = "none"


class UpDown(Enum):
    """
    The up-down type is used for the direction of arrows and other pointed symbols
    like vertical accents, indicating which way the tip is pointing.
    """

    UP = "up"
    DOWN = "down"


class UpDownStopContinue(Enum):
    """
    The up-down-stop-continue type is used for octave-shift elements, indicating
    the direction of the shift from their true pitched values because of printing
    difficulty.
    """

    UP = "up"
    DOWN = "down"
    STOP = "stop"
    CONTINUE = "continue"


class UprightInverted(Enum):
    """The upright-inverted type describes the appearance of a fermata element.

    The value is upright if not specified.
    """

    UPRIGHT = "upright"
    INVERTED = "inverted"


class Valign(Enum):
    """The valign type is used to indicate vertical alignment to the top, middle,
    bottom, or baseline of the text.

    If the text is on multiple lines, baseline alignment refers to the
    baseline of the lowest line of text. Defaults are implementation-
    dependent.
    """

    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"
    BASELINE = "baseline"


class ValignImage(Enum):
    """The valign-image type is used to indicate vertical alignment for images and
    graphics, so it does not include a baseline value.

    Defaults are implementation-dependent.
    """

    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"


class WedgeType(Enum):
    """The wedge type is crescendo for the start of a wedge that is closed at the
    left side, diminuendo for the start of a wedge that is closed on the right
    side, and stop for the end of a wedge.

    The continue type is used for formatting wedges over a system break,
    or for other situations where a single wedge is divided into
    multiple segments.
    """

    CRESCENDO = "crescendo"
    DIMINUENDO = "diminuendo"
    STOP = "stop"
    CONTINUE = "continue"


class Winged(Enum):
    """The winged attribute indicates whether the repeat has winged extensions that
    appear above and below the barline.

    The straight and curved values represent single wings, while the
    double-straight and double-curved values represent double wings. The
    none value indicates no wings and is the default.
    """

    NONE = "none"
    STRAIGHT = "straight"
    CURVED = "curved"
    DOUBLE_STRAIGHT = "double-straight"
    DOUBLE_CURVED = "double-curved"


class WoodValue(Enum):
    """The wood-value type represents pictograms for wood percussion instruments.

    The maraca and maracas values distinguish the one- and two-maraca
    versions of the pictogram.
    """

    BAMBOO_SCRAPER = "bamboo scraper"
    BOARD_CLAPPER = "board clapper"
    CABASA = "cabasa"
    CASTANETS = "castanets"
    CASTANETS_WITH_HANDLE = "castanets with handle"
    CLAVES = "claves"
    FOOTBALL_RATTLE = "football rattle"
    GUIRO = "guiro"
    LOG_DRUM = "log drum"
    MARACA = "maraca"
    MARACAS = "maracas"
    QUIJADA = "quijada"
    RAINSTICK = "rainstick"
    RATCHET = "ratchet"
    RECO_RECO = "reco-reco"
    SANDPAPER_BLOCKS = "sandpaper blocks"
    SLIT_DRUM = "slit drum"
    TEMPLE_BLOCK = "temple block"
    VIBRASLAP = "vibraslap"
    WHIP = "whip"
    WOOD_BLOCK = "wood block"


class YesNo(Enum):
    """The yes-no type is used for boolean-like attributes.

    We cannot use W3C XML Schema booleans due to their restrictions on
    expression of boolean values.
    """

    YES = "yes"
    NO = "no"



from dataclasses import InitVar, dataclass, field

from libresvip.core.tick_counter import skip_tempo_list
from libresvip.model.base import SongTempo
from libresvip.utils.search import find_last_index


@dataclass
class TimeSynchronizer:
    is_absolute_time_code: bool = field(init=False)
    default_tempo: float = field(init=False)
    tempo_list: list[SongTempo] = field(init=False)
    ori_tempo_list: InitVar[list[SongTempo]]
    skip_ticks: InitVar[int] = 0
    _is_absolute_time_code: InitVar[bool] = False
    _default_tempo: InitVar[float] = 60

    def __post_init__(
        self,
        ori_tempo_list: list[SongTempo],
        skip_ticks: int,
        _is_absolute_time_code: bool,
        _default_tempo: float,
    ) -> None:
        if skip_ticks > 0:
            self.tempo_list = skip_tempo_list(ori_tempo_list, skip_ticks)
        else:
            self.tempo_list = ori_tempo_list
        self.is_absolute_time_code = _is_absolute_time_code
        self.default_tempo = _default_tempo

    def get_actual_ticks_from_ticks(self, ticks: int) -> float:
        if not self.is_absolute_time_code:
            return ticks
        res = 0.0
        i = 0
        while i < len(self.tempo_list) - 1 and self.tempo_list[i + 1].position < ticks:
            res += (self.tempo_list[i + 1].position - self.tempo_list[i].position) * (
                self.default_tempo / self.tempo_list[i].bpm
            )
            i += 1
        res += (ticks - self.tempo_list[i].position) * self.default_tempo / self.tempo_list[i].bpm
        return res

    def get_duration_secs_from_ticks(self, start_ticks: int, end_ticks: int) -> float:
        if self.is_absolute_time_code:
            return (
                (
                    self.get_actual_ticks_from_ticks(end_ticks)
                    - self.get_actual_ticks_from_ticks(start_ticks)
                )
                / self.default_tempo
                / 8
            )
        start_tempo_index = find_last_index(
            self.tempo_list, lambda tempo: tempo.position <= start_ticks
        )
        end_tempo_index = find_last_index(
            self.tempo_list, lambda tempo: tempo.position <= end_ticks
        )
        if start_tempo_index == end_tempo_index or end_tempo_index == -1:
            return (end_ticks - start_ticks) / self.tempo_list[start_tempo_index].bpm / 8
        secs = 0.0
        secs += (
            (self.tempo_list[start_tempo_index + 1].position - start_ticks)
            / self.tempo_list[start_tempo_index].bpm
            / 8
        )
        for i in range(start_tempo_index + 1, end_tempo_index):
            secs += (
                (self.tempo_list[i + 1].position - self.tempo_list[i].position)
                / float(self.tempo_list[i].bpm)
                / 8
            )
        secs += (
            (end_ticks - self.tempo_list[end_tempo_index].position)
            / float(self.tempo_list[end_tempo_index].bpm)
            / 8
        )
        return secs

    def get_actual_ticks_from_secs_offset(self, start_ticks: int, offset_secs: float) -> float:
        if self.is_absolute_time_code:
            return (
                self.get_actual_ticks_from_ticks(start_ticks) + offset_secs * self.default_tempo * 8
            )
        start_tempo_index = find_last_index(
            self.tempo_list, lambda tempo: tempo.position <= start_ticks
        )
        ticks: float = start_ticks
        secs = offset_secs
        for i in range(start_tempo_index, len(self.tempo_list) - 1):
            dur = (self.tempo_list[i + 1].position - ticks) / self.tempo_list[i].bpm / 8
            if dur < secs:
                ticks = self.tempo_list[i + 1].position
                secs -= dur
            else:
                ticks += (self.tempo_list[i + 1].position - ticks) * secs / dur
                return ticks
        ticks += self.tempo_list[-1].bpm * secs * 8
        return ticks

    def get_actual_ticks_from_secs(self, secs: float) -> float:
        return self.get_actual_ticks_from_secs_offset(0, secs)

    def get_actual_secs_from_ticks(self, ticks: int) -> float:
        return self.get_duration_secs_from_ticks(0, ticks)



from collections.abc import Callable
from operator import attrgetter

from more_itertools import minmax

from libresvip.model.point import Point
from libresvip.utils.search import find_index, find_last_index


def resampled(
    data: list[Point],
    interval: int,
    interpolate_method: Callable[[Point, Point, int], float],
) -> list[Point]:
    left_point, right_point = minmax(data, key=attrgetter("x"), default=(Point(0, 0), Point(0, 0)))
    return [
        Point(
            x=current,
            y=int(interpolate_method(data[prev_index], data[next_index], current)),
        )
        for current in range(left_point.x, right_point.x + 1, interval)
        if (prev_index := find_last_index(data, lambda p: p.x <= current)) != -1
        and (next_index := find_index(data, lambda p: p.x >= current)) != -1
    ]


def dot_resampled(data: list[Point], interval: int) -> list[Point]:
    return resampled(
        data,
        interval,
        lambda prev, next_, _: prev[1] if prev and prev[1] is not None else next_[1],
    )



from enum import Enum

from pydantic import Field

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import BaseModel


class GjgjBeatStyle(Enum):
    NONE = 0
    SP = 1
    SIL = 2


class GjgjBeatItems(BaseModel):
    id_value: int | None = Field(alias="ID")
    lyric: str = Field(alias="Lyric")
    pinyin: str = Field(alias="Pinyin")
    start_tick: int = Field(alias="StartTick")
    duration: int = Field(alias="Duration")
    track: int | None = Field(alias="Track")
    pre_time: float = Field(0.0, alias="PreTime")
    post_time: float = Field(0.0, alias="PostTime")
    style: GjgjBeatStyle | None = Field(alias="Style")
    velocity: int | None = Field(127, alias="Velocity")


class GjgjKeyboard(BaseModel):
    key_mode: int | None = Field(1, alias="KeyMode")
    key_type: int | None = Field(0, alias="KeyType")


class GjgjTrackVolume(BaseModel):
    volume: float | None = Field(1.0, alias="Volume")
    left_volume: float | None = Field(1.0, alias="LeftVolume")
    right_volume: float | None = Field(1.0, alias="RightVolume")
    mute: bool = Field(False, alias="Mute")


class GjgjProjectSetting(BaseModel):
    no1_key_name: str | None = Field("C", alias="No1KeyName")
    eq_after_mix: str | None = Field("", alias="EQAfterMix")
    project_type: int | None = Field(0, alias="ProjectType")
    denominator: int | None = Field(4, alias="Denominator")
    syn_mode: int | None = Field(0, alias="SynMode")


class GjgjSingerInfo(BaseModel):
    name: str = Field("", alias="Name")
    display_name: str = Field("", alias="DisplayName")
    template_id: str | None = Field("", alias="TemplateID")
    sex: str | None = Field("", alias="Sex")
    age: int | None = Field(100, alias="Age")
    color: int | None = Field(0, alias="Color")
    user_id: str | None = Field("", alias="UserID")
    up_pitch: int | None = Field(0, alias="UpPitch")
    low_pitch: int | None = Field(0, alias="LowPitch")
    recommend_up_pitch: int | None = Field(0, alias="RecommandUpPitch")
    recommend_low_pitch: int | None = Field(0, alias="RecommandLowPitch")
    image: str | None = ""
    image_url: str | None = ""
    full_avatar: str | None = ""
    synthesize_file: str | None = Field("", alias="synthetize_file")


class GjgjTempos(BaseModel):
    time: int | None = Field(alias="Time")
    microseconds_per_quarter_note: int | None = Field(alias="MicrosecondsPerQuarterNote")


class GjgjTimeSignature(BaseModel):
    time: int = Field(0, alias="Time")
    numerator: int = Field(4, alias="Numerator")
    denominator: int = Field(4, alias="Denominator")


class GjgjPoint(BaseModel):
    x: float = Field(alias="X")
    y: float = Field(alias="Y")


class GjgjTone(BaseModel):
    modifies: list[GjgjPoint] = Field(default_factory=list, alias="Modifys")
    modify_ranges: list[GjgjPoint] = Field(default_factory=list, alias="ModifyRanges")


class GjgjTempoMap(BaseModel):
    ticks_per_quarter_note: int = Field(TICKS_IN_BEAT, alias="TicksPerQuarterNote")
    tempos: list[GjgjTempos] = Field(default_factory=list, alias="Tempos")
    time_signature: list[GjgjTimeSignature] = Field(default_factory=list, alias="TimeSignature")


class GjgjVolumeMap(BaseModel):
    time: int = Field(0, alias="Time")
    volume: float = Field(0.0, alias="Volume")


class GjgjSingingTrack(BaseModel):
    id_value: str | None = Field(None, alias="ID")
    type_: int | None = Field(0, alias="Type")
    name: str | None = Field(None, alias="Name")
    beat_items: list[GjgjBeatItems] = Field(default_factory=list, alias="BeatItems")
    tone: GjgjTone = Field(default_factory=GjgjTone, alias="Tone")
    volume_map: list[GjgjVolumeMap] = Field(default_factory=list, alias="VolumeMap")
    singer_info: GjgjSingerInfo = Field(default_factory=GjgjSingerInfo, alias="SingerInfo")
    keyboard: GjgjKeyboard | None = Field(default_factory=GjgjKeyboard, alias="Keyboard")
    master_volume: GjgjTrackVolume | None = Field(None, alias="MasterVolume")
    eq_program: str | None = Field("无", alias="EQProgram")
    sort_index: int | None = Field(0, alias="SortIndex")


class GjgjInstrumentalTrack(BaseModel):
    id_value: str | None = Field(None, alias="ID")
    type_: int | None = Field(1, alias="Type")
    path: str | None = Field(None, alias="Path")
    offset: int = Field(alias="Offset")
    master_volume: GjgjTrackVolume | None = Field(None, alias="MasterVolume")
    eq_program: str | None = Field("", alias="EQProgram")
    sort_index: int | None = Field(0, alias="SortIndex")


class GjgjMidiControl(BaseModel):
    control_num: int | None = Field(None, alias="ControlNum")
    volume_map: list[GjgjVolumeMap] = Field(default_factory=list, alias="VolumeMap")


class GjgjMidiTrack(BaseModel):
    id_value: str | None = Field(None, alias="ID")
    type_: int | None = Field(2, alias="Type")
    sort_index: int | None = Field(None, alias="SortIndex")
    bank: int | None = Field(None, alias="Bank")
    program: int | None = Field(None, alias="Program")
    channel: int | None = Field(None, alias="Channel")
    beat_items: list[GjgjBeatItems] = Field(default_factory=list, alias="BeatItems")
    control_map: list[GjgjMidiControl] = Field(default_factory=list, alias="ControlMap")


class GjgjProject(BaseModel):
    gjgj_version: int | None = Field(2, alias="gjgjVersion")
    project_setting: GjgjProjectSetting | None = Field(
        default_factory=GjgjProjectSetting, alias="ProjectSetting"
    )
    accompaniments: list[GjgjInstrumentalTrack] = Field(
        default_factory=list, alias="Accompaniments"
    )
    tracks: list[GjgjSingingTrack] = Field(default_factory=list, alias="Tracks")
    midi_tracks: list[GjgjMidiTrack] = Field(default_factory=list, alias="MIDITracks")
    tempo_map: GjgjTempoMap = Field(default_factory=GjgjTempoMap, alias="TempoMap")



import dataclasses
import re

import pypinyin

from libresvip.core.constants import DEFAULT_BPM
from libresvip.core.exceptions import NotesOverlappedError
from libresvip.core.lyric_phoneme.chinese import CHINESE_RE
from libresvip.core.lyric_phoneme.japanese import is_kana, to_romaji
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import db_to_float
from libresvip.utils.translation import gettext_lazy as _

from .model import (
    UNote,
    USTXProject,
    UTempo,
    UTimeSignature,
    UTrack,
    UVoicePart,
    UWavePart,
)
from .options import InputOptions, OpenUtauEnglishPhonemizerCompatibility
from .util import BasePitchGenerator

PHONETIC_HINT_RE = re.compile(r"\[(.*?)\]")


@dataclasses.dataclass
class UstxParser:
    options: InputOptions
    base_pitch_generator: BasePitchGenerator = dataclasses.field(init=False)

    def parse_project(self, ustx_project: USTXProject) -> Project:
        self.breath_lyrics = self.options.breath_lyrics.strip().split()
        self.silence_lyrics = self.options.silence_lyrics.strip().split()
        self.base_pitch_generator = BasePitchGenerator(ustx_project)
        tempos = self.parse_tempos(ustx_project.tempos)
        time_signatures = self.parse_time_signatures(ustx_project.time_signatures)
        tracks = self.parse_tracks(ustx_project.tracks, ustx_project.voice_parts)
        for track in tracks:
            track.edited_params.pitch.points.append(Point.end_point())
        tracks.extend(self.parse_wave_parts(ustx_project.tracks, ustx_project.wave_parts))
        return Project(
            song_tempo_list=tempos,
            time_signature_list=time_signatures,
            track_list=tracks,
        )

    @staticmethod
    def parse_tempos(tempos: list[UTempo]) -> list[SongTempo]:
        song_tempo_list = [
            SongTempo(
                position=tempo.position + 1920 if tempo.position > 0 else tempo.position,
                bpm=tempo.bpm,
            )
            for tempo in tempos
        ]
        if not song_tempo_list:
            song_tempo_list.append(SongTempo(position=0, bpm=DEFAULT_BPM))
        return song_tempo_list

    @staticmethod
    def parse_time_signatures(
        time_signatures: list[UTimeSignature],
    ) -> list[TimeSignature]:
        time_signature_list = [
            TimeSignature(
                bar_index=time_signature.bar_position,
                numerator=time_signature.beat_per_bar,
                denominator=time_signature.beat_unit,
            )
            for time_signature in time_signatures
        ]
        if not time_signature_list:
            time_signature_list.append(TimeSignature())
        return time_signature_list

    def parse_tracks(self, tracks: list[UTrack], voice_parts: list[UVoicePart]) -> list[Track]:
        track_list = [
            SingingTrack(
                volume=self.parse_volume(ustx_track.volume),
                solo=ustx_track.solo,
                mute=ustx_track.mute,
                ai_singer_name=ustx_track.singer or "",
                title=ustx_track.track_name or f"Track {i + 1}",
            )
            for i, ustx_track in enumerate(tracks)
        ]
        if self.options.import_pitch:
            for track in track_list:
                track.edited_params.pitch.points.append(Point.start_point())
        for voice_part in voice_parts:
            track_index = voice_part.track_no
            if track_index < len(track_list):
                singing_track: SingingTrack = track_list[track_index]
            else:
                continue
            if not singing_track.title:
                singing_track.title = voice_part.name
            notes = self.parse_notes(voice_part.notes, voice_part.position)
            singing_track.note_list.extend(notes)
            if self.options.import_pitch:
                singing_track.edited_params.pitch.points.root.extend(self.parse_pitch(voice_part))
        return [track for track in track_list if len(track.note_list)]

    def parse_pitch(self, part: UVoicePart) -> list[Point]:
        pitch_start = self.base_pitch_generator.pitch_start
        pitch_interval = self.base_pitch_generator.pitch_interval
        first_bar_length = 1920

        pitches = self.base_pitch_generator.base_pitch(part)

        curve = next((c for c in part.curves if c.abbr == "pitd"), None)
        if curve is not None and not curve.is_empty:
            for i in range(len(pitches)):
                pitches[i] += curve.sample(pitch_start + i * pitch_interval)

        point_list = [Point(first_bar_length + part.position, -100)]
        point_list.extend(
            Point(
                first_bar_length + part.position + i * pitch_interval,
                int(pitches[i]),
            )
            for i in range(len(pitches))
        )
        point_list.append(
            Point(
                first_bar_length + part.position + len(pitches) * pitch_interval,
                -100,
            )
        )
        return point_list

    def parse_notes(self, notes: list[UNote], tick_prefix: int) -> list[Note]:
        note_list = []
        prev_ustx_note = None
        for ustx_note in notes:
            note_lyric = ustx_note.lyric
            if note_lyric.startswith("+"):
                if (
                    note_lyric.removeprefix("+").isdigit()
                    and self.options.english_phonemizer_compatibility
                    == OpenUtauEnglishPhonemizerCompatibility.ARPA
                ):
                    note_lyric = "+"
                else:
                    note_lyric = "-"
            note = Note(
                key_number=ustx_note.tone,
                lyric=note_lyric,
                start_pos=ustx_note.position + tick_prefix,
                length=ustx_note.duration,
            )
            if ustx_note.lyric:
                if (phonetic_hint := PHONETIC_HINT_RE.search(ustx_note.lyric)) is not None:
                    note.pronunciation = phonetic_hint.group(1)
                elif is_kana(ustx_note.lyric):
                    note.pronunciation = to_romaji(ustx_note.lyric)
                elif (chinese_char := CHINESE_RE.search(ustx_note.lyric)) is not None:
                    note.pronunciation = " ".join(pypinyin.lazy_pinyin(chinese_char.group()))
                elif not ustx_note.lyric.startswith("+"):
                    note.pronunciation = ustx_note.lyric.removeprefix("?")
            if prev_ustx_note is not None:
                if prev_ustx_note.end > ustx_note.position:
                    msg = _("Notes Overlapped")
                    raise NotesOverlappedError(msg)
                elif (
                    prev_ustx_note.end == ustx_note.position
                    and prev_ustx_note.lyric in self.breath_lyrics
                ):
                    note.head_tag = "V"
            if (
                ustx_note.lyric not in self.breath_lyrics
                and ustx_note.lyric not in self.silence_lyrics
            ):
                note_list.append(note)
            prev_ustx_note = ustx_note
        return note_list

    def parse_wave_parts(
        self, tracks: list[UTrack], wave_parts: list[UWavePart]
    ) -> list[InstrumentalTrack]:
        track_list = []
        if self.options.import_instrumental_track:
            for wave_part in wave_parts:
                ustx_track = tracks[wave_part.track_no]
                # duration = wave_part.file_duration_ms - wave_part.skip_ms - wave_part.trim_ms
                track_list.append(
                    InstrumentalTrack(
                        audio_file_path=wave_part.relative_path,
                        offset=wave_part.position,
                        title=wave_part.name,
                        mute=ustx_track.mute,
                        solo=ustx_track.solo,
                        volume=self.parse_volume(ustx_track.volume),
                    )
                )
        return track_list

    def parse_volume(self, volume: float) -> float:
        return min(db_to_float(volume, using_amplitude=False), 2)



import io
import pathlib

import mido_fix as mido

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .options import InputOptions, OutputOptions
from .vsq_generator import VsqGenerator
from .vsq_parser import VsqParser


class VocaloidSequenceConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        vsq_file = mido.MidiFile(
            file=io.BytesIO(path.read_bytes()),
            charset=options.lyric_encoding,
            clip=True,
        )
        return VsqParser(
            options=options,
        ).parse_project(vsq_file)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        vsq_file = VsqGenerator(
            options=options,
        ).generate_project(project)
        buffer = io.BytesIO()
        vsq_file.save(file=buffer)
        path.write_bytes(buffer.getvalue())



import contextlib


def ansi2unicode(content: bytes) -> str:
    for possible_encoding in [
        "utf-8",
        "gbk",
        "shift-jis",
        "euc-kr",
        "big5",
    ]:
        with contextlib.suppress(UnicodeDecodeError):
            return content.decode(possible_encoding)
    return content.decode(errors="replace")



from enum import Enum
from typing import Optional

from xsdata_pydantic.fields import field

from libresvip.model.base import BaseModel

from .enums import VocaloidLanguage

VSQ3_NS = "http://www.yamaha.co.jp/vocaloid/schema/vsq3/"


class ParameterNames(Enum):
    DYN = "DYN"
    BRE = "BRE"
    BRI = "BRI"
    CLE = "CLE"
    GEN = "GEN"
    POR = "POR"
    PBS = "PBS"
    PIT = "PIT"


class Aux(BaseModel):
    class Meta:
        name = "aux"
        namespace = VSQ3_NS

    aux_id: str | None = field(
        default=None,
        metadata={
            "name": "auxID",
            "type": "Element",
            "required": True,
        },
    )
    content: str | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Dyn(BaseModel):
    class Meta:
        name = "dyn"
        namespace = VSQ3_NS

    pos_sec: int | None = field(
        default=None,
        metadata={
            "name": "posSec",
            "type": "Element",
            "required": True,
        },
    )
    dyn_val: int | None = field(
        default=None,
        metadata={
            "name": "dynVal",
            "type": "Element",
            "required": True,
        },
    )


class Pch(BaseModel):
    class Meta:
        name = "pch"
        namespace = VSQ3_NS

    pos_sec: int | None = field(
        default=None,
        metadata={
            "name": "posSec",
            "type": "Element",
            "required": True,
        },
    )
    pch_val: int | None = field(
        default=None,
        metadata={
            "name": "pchVal",
            "type": "Element",
            "required": True,
        },
    )


class Phnm(BaseModel):
    class Meta:
        name = "phnm"
        namespace = VSQ3_NS

    pos_sec: int | None = field(
        default=None,
        metadata={
            "name": "posSec",
            "type": "Element",
            "required": True,
        },
    )
    dur_sec: int | None = field(
        default=None,
        metadata={
            "name": "durSec",
            "type": "Element",
            "required": True,
        },
    )
    phnm_str: str | None = field(
        default=None,
        metadata={
            "name": "phnmStr",
            "type": "Element",
            "required": True,
        },
    )


class SeqAttr(BaseModel):
    class Meta:
        name = "seqAttr"
        namespace = VSQ3_NS

    elem: list["SeqAttr.Elem"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    seq_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Attribute",
            "required": True,
        },
    )

    class Elem(BaseModel):
        pos_nrm: int | None = field(
            default=None,
            metadata={
                "name": "posNrm",
                "type": "Element",
                "required": True,
            },
        )
        elv: int | None = field(
            default=None,
            metadata={
                "type": "Element",
                "required": True,
            },
        )


class Singer(BaseModel):
    class Meta:
        name = "singer"
        namespace = VSQ3_NS

    pos_tick: int | None = field(
        default=0,
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    v_bs: VocaloidLanguage | None = field(
        default=VocaloidLanguage.SIMPLIFIED_CHINESE,
        metadata={
            "name": "vBS",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    v_pc: int | None = field(
        default=0,
        metadata={
            "name": "vPC",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )


class StylePlugin(BaseModel):
    class Meta:
        name = "stylePlugin"
        namespace = VSQ3_NS

    style_plugin_id: str | None = field(
        default="ACA9C502-A04B-42b5-B2EB-5CEA36D16FCE",
        metadata={
            "name": "stylePluginID",
            "type": "Element",
            "required": True,
            "length": 36,
        },
    )
    style_plugin_name: str | None = field(
        default="VOCALOID2 Compatible Style",
        metadata={
            "name": "stylePluginName",
            "type": "Element",
            "required": True,
        },
    )
    version: str | None = field(
        default="3.0.0.1",
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Tempo(BaseModel):
    class Meta:
        name = "tempo"
        namespace = VSQ3_NS

    pos_tick: int = field(
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    bpm: int = field(
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class TimeSig(BaseModel):
    class Meta:
        name = "timeSig"
        namespace = VSQ3_NS

    pos_mes: int | None = field(
        default=None,
        metadata={
            "name": "posMes",
            "type": "Element",
            "required": True,
        },
    )
    nume: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    denomi: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class TypeParamAttr(BaseModel):
    class Meta:
        name = "typeParamAttr"

    value: int | None = field(
        default=None,
        metadata={
            "required": True,
        },
    )
    type_param_attr_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Attribute",
            "required": True,
        },
    )


class TypePhonemes(BaseModel):
    class Meta:
        name = "typePhonemes"

    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )
    lock: int | None = field(
        default=None,
        metadata={
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )


class TypeResolution(Enum):
    VALUE_16 = 16
    VALUE_24 = 24
    VALUE_32 = 32


class TypeSamplerate(Enum):
    VALUE_44100 = 44100
    VALUE_48000 = 48000
    VALUE_96000 = 96000


class TypeVstSdkversion(Enum):
    VALUE_0 = 0
    VALUE_2 = 2
    VALUE_3 = 3


class VVoiceParam(BaseModel):
    class Meta:
        name = "vVoiceParam"
        namespace = VSQ3_NS

    bre: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    bri: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    cle: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    gen: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    ope: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )


class Voice(BaseModel):
    class Meta:
        name = "voice"
        namespace = VSQ3_NS

    pos_sec: int | None = field(
        default=None,
        metadata={
            "name": "posSec",
            "type": "Element",
            "required": True,
        },
    )
    v_bs: VocaloidLanguage | None = field(
        default=VocaloidLanguage.SIMPLIFIED_CHINESE,
        metadata={
            "name": "vBS",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    v_pc: int | None = field(
        default=0,
        metadata={
            "name": "vPC",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )


class MCtrl(BaseModel):
    class Meta:
        name = "mCtrl"
        namespace = VSQ3_NS

    pos_tick: int = field(
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    attr: TypeParamAttr = field(
        default_factory=TypeParamAttr,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class MasterTrack(BaseModel):
    class Meta:
        name = "masterTrack"
        namespace = VSQ3_NS

    seq_name: str | None = field(
        default="Untitled0",
        metadata={
            "name": "seqName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="New VSQ File",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    resolution: int | None = field(
        default=480,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    pre_measure: int = field(
        default=1,
        metadata={
            "name": "preMeasure",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    time_sig: list[TimeSig] = field(
        default_factory=list,
        metadata={
            "name": "timeSig",
            "type": "Element",
            "min_occurs": 1,
        },
    )
    tempo: list[Tempo] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )


class NoteStyle(BaseModel):
    class Meta:
        name = "noteStyle"
        namespace = VSQ3_NS

    attr: list[TypeParamAttr] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )
    seq_attr: list[SeqAttr] = field(
        default_factory=list,
        metadata={
            "name": "seqAttr",
            "type": "Element",
        },
    )


class PCtrl(BaseModel):
    class Meta:
        name = "pCtrl"
        namespace = VSQ3_NS

    pos_sec: int | None = field(
        default=None,
        metadata={
            "name": "posSec",
            "type": "Element",
            "required": True,
        },
    )
    attr: TypeParamAttr | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class PartStyle(BaseModel):
    class Meta:
        name = "partStyle"
        namespace = VSQ3_NS

    attr: list[TypeParamAttr] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )


class PhraseStyle(BaseModel):
    class Meta:
        name = "phraseStyle"
        namespace = VSQ3_NS

    pos_tick: int | None = field(
        default=None,
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    attr: list[TypeParamAttr] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )


class SCtrl(BaseModel):
    class Meta:
        name = "sCtrl"
        namespace = VSQ3_NS

    pos_tick: int | None = field(
        default=None,
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    attr: TypeParamAttr | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class VVoice(BaseModel):
    class Meta:
        name = "vVoice"
        namespace = VSQ3_NS

    v_bs: VocaloidLanguage | None = field(
        default=VocaloidLanguage.SIMPLIFIED_CHINESE,
        metadata={
            "name": "vBS",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    v_pc: int | None = field(
        default=0,
        metadata={
            "name": "vPC",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    comp_id: str | None = field(
        default=None,
        metadata={
            "name": "compID",
            "type": "Element",
            "required": True,
            "length": 16,
        },
    )
    v_voice_name: str | None = field(
        default=None,
        metadata={
            "name": "vVoiceName",
            "type": "Element",
            "required": True,
        },
    )
    v_voice_param: VVoiceParam | None = field(
        default_factory=VVoiceParam,
        metadata={
            "name": "vVoiceParam",
            "type": "Element",
            "required": True,
        },
    )


class VstPlugin(BaseModel):
    class Meta:
        name = "vstPlugin"
        namespace = VSQ3_NS

    vst_plugin_id: str | None = field(
        default=None,
        metadata={
            "name": "vstPluginID",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin_name: str | None = field(
        default=None,
        metadata={
            "name": "vstPluginName",
            "type": "Element",
            "required": True,
        },
    )
    vst_sdkversion: TypeVstSdkversion | None = field(
        default=None,
        metadata={
            "name": "vstSDKVersion",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_num: int | None = field(
        default=None,
        metadata={
            "name": "vstParamNum",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_val: Optional["VstPlugin.VstParamVal"] = field(
        default=None,
        metadata={
            "name": "vstParamVal",
            "type": "Element",
        },
    )
    vst_preset_no: int | None = field(
        default=None,
        metadata={
            "name": "vstPresetNo",
            "type": "Element",
        },
    )
    enable: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    bypass: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )

    class VstParamVal(BaseModel):
        val: list[int] = field(
            default_factory=list,
            metadata={
                "type": "Element",
            },
        )


class VstPluginSr(BaseModel):
    class Meta:
        name = "vstPluginSR"
        namespace = VSQ3_NS

    vst_plugin_id: str | None = field(
        default=None,
        metadata={
            "name": "vstPluginID",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin_name: str | None = field(
        default=None,
        metadata={
            "name": "vstPluginName",
            "type": "Element",
            "required": True,
        },
    )
    vst_sdkversion: TypeVstSdkversion | None = field(
        default=None,
        metadata={
            "name": "vstSDKVersion",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_num: int | None = field(
        default=None,
        metadata={
            "name": "vstParamNum",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_val: Optional["VstPluginSr.VstParamVal"] = field(
        default=None,
        metadata={
            "name": "vstParamVal",
            "type": "Element",
        },
    )
    vst_preset_no: int | None = field(
        default=None,
        metadata={
            "name": "vstPresetNo",
            "type": "Element",
        },
    )
    enable: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    bypass: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )

    class VstParamVal(BaseModel):
        val: list[int] = field(
            default_factory=list,
            metadata={
                "type": "Element",
            },
        )


class WavPart(BaseModel):
    class Meta:
        name = "wavPart"
        namespace = VSQ3_NS

    pos_tick: int | None = field(
        default=None,
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    play_time: int | None = field(
        default=None,
        metadata={
            "name": "playTime",
            "type": "Element",
            "required": True,
        },
    )
    part_name: str | None = field(
        default=None,
        metadata={
            "name": "partName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    sample_rate: TypeSamplerate | None = field(
        default=None,
        metadata={
            "name": "sampleRate",
            "type": "Element",
            "required": True,
        },
    )
    sample_reso: TypeResolution | None = field(
        default=None,
        metadata={
            "name": "sampleReso",
            "type": "Element",
            "required": True,
        },
    )
    channels: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 1,
            "max_inclusive": 2,
        },
    )
    file_path: str | None = field(
        default=None,
        metadata={
            "name": "filePath",
            "type": "Element",
            "required": True,
        },
    )


class StereoTrack(BaseModel):
    class Meta:
        name = "karaokeTrack"
        namespace = VSQ3_NS

    wav_part: list[WavPart] = field(
        default_factory=list,
        metadata={
            "name": "wavPart",
            "type": "Element",
        },
    )


class StereoUnit(BaseModel):
    class Meta:
        name = "karaokeUnit"
        namespace = VSQ3_NS

    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "inGain",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "vstPlugin",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class MasterUnit(BaseModel):
    class Meta:
        name = "masterUnit"
        namespace = VSQ3_NS

    out_dev: int | None = field(
        default=0,
        metadata={
            "name": "outDev",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "vstPlugin",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    vst_plugin_sr: VstPluginSr | None = field(
        default=None,
        metadata={
            "name": "vstPluginSR",
            "type": "Element",
        },
    )
    ret_level: int | None = field(
        default=0,
        metadata={
            "name": "retLevel",
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Note(BaseModel):
    class Meta:
        name = "note"
        namespace = VSQ3_NS

    pos_tick: int = field(
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    dur_tick: int = field(
        metadata={
            "name": "durTick",
            "type": "Element",
            "required": True,
        },
    )
    note_num: int | None = field(
        default=None,
        metadata={
            "name": "noteNum",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    velocity: int | None = field(
        default=64,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    lyric: str = field(
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    phnms: TypePhonemes | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    note_style: NoteStyle = field(
        default_factory=NoteStyle,
        metadata={
            "name": "noteStyle",
            "type": "Element",
        },
    )


class ProsodyPart(BaseModel):
    class Meta:
        name = "prosodyPart"
        namespace = VSQ3_NS

    pos_tick: int | None = field(
        default=None,
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    play_time: int | None = field(
        default=None,
        metadata={
            "name": "playTime",
            "type": "Element",
            "required": True,
        },
    )
    part_name: str | None = field(
        default=None,
        metadata={
            "name": "partName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    voice: list[Voice] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    p_ctrl: list[PCtrl] = field(
        default_factory=list,
        metadata={
            "name": "pCtrl",
            "type": "Element",
        },
    )
    pch: list[Pch] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    dyn: list[Dyn] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    phnm: list[Phnm] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )


class MonoTrack(BaseModel):
    class Meta:
        name = "seTrack"
        namespace = VSQ3_NS

    wav_part: list[WavPart] = field(
        default_factory=list,
        metadata={
            "name": "wavPart",
            "type": "Element",
        },
    )


class MonoUnit(BaseModel):
    class Meta:
        name = "seUnit"
        namespace = VSQ3_NS

    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "inGain",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "vstPlugin",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    send_level: int | None = field(
        default=-898,
        metadata={
            "name": "sendLevel",
            "type": "Element",
            "required": True,
        },
    )
    send_enable: int | None = field(
        default=0,
        metadata={
            "name": "sendEnable",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    pan: int | None = field(
        default=64,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class VVoiceTable(BaseModel):
    class Meta:
        name = "vVoiceTable"
        namespace = VSQ3_NS

    v_voice: list[VVoice] = field(
        default_factory=list,
        metadata={
            "name": "vVoice",
            "type": "Element",
            "min_occurs": 1,
        },
    )


class VsUnit(BaseModel):
    class Meta:
        name = "vsUnit"
        namespace = VSQ3_NS

    vs_track_no: int | None = field(
        default=None,
        metadata={
            "name": "vsTrackNo",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "inGain",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "vstPlugin",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    send_level: int | None = field(
        default=-898,
        metadata={
            "name": "sendLevel",
            "type": "Element",
            "required": True,
        },
    )
    send_enable: int | None = field(
        default=0,
        metadata={
            "name": "sendEnable",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    pan: int | None = field(
        default=64,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Mixer(BaseModel):
    class Meta:
        name = "mixer"
        namespace = VSQ3_NS

    master_unit: MasterUnit | None = field(
        default_factory=MasterUnit,
        metadata={
            "name": "masterUnit",
            "type": "Element",
            "required": True,
        },
    )
    vs_unit: list[VsUnit] = field(
        default_factory=list,
        metadata={
            "name": "vsUnit",
            "type": "Element",
            "min_occurs": 1,
            "max_occurs": 16,
        },
    )
    mono_unit: MonoUnit = field(
        default_factory=MonoUnit,
        metadata={
            "name": "seUnit",
            "type": "Element",
            "required": True,
        },
    )
    stereo_unit: StereoUnit = field(
        default_factory=StereoUnit,
        metadata={
            "name": "karaokeUnit",
            "type": "Element",
            "required": True,
        },
    )


class MusicalPart(BaseModel):
    class Meta:
        name = "musicalPart"
        namespace = VSQ3_NS

    pos_tick: int = field(
        metadata={
            "name": "posTick",
            "type": "Element",
            "required": True,
        },
    )
    play_time: int = field(
        metadata={
            "name": "playTime",
            "type": "Element",
            "required": True,
        },
    )
    part_name: str | None = field(
        default="New Part",
        metadata={
            "name": "partName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="New Musical Part",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    style_plugin: StylePlugin | None = field(
        default_factory=StylePlugin,
        metadata={
            "name": "stylePlugin",
            "type": "Element",
            "required": True,
        },
    )
    part_style: PartStyle = field(
        default_factory=PartStyle,
        metadata={
            "name": "partStyle",
            "type": "Element",
        },
    )
    phrase_style: list[PhraseStyle] = field(
        default_factory=list,
        metadata={
            "name": "phraseStyle",
            "type": "Element",
        },
    )
    singer: list[Singer] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    m_ctrl: list[MCtrl] = field(
        default_factory=list,
        metadata={
            "name": "mCtrl",
            "type": "Element",
        },
    )
    s_ctrl: list[SCtrl] = field(
        default_factory=list,
        metadata={
            "name": "sCtrl",
            "type": "Element",
        },
    )
    note: list[Note] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )


class VsTrack(BaseModel):
    class Meta:
        name = "vsTrack"
        namespace = VSQ3_NS

    vs_track_no: int | None = field(
        default=None,
        metadata={
            "name": "vsTrackNo",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    track_name: str | None = field(
        default=None,
        metadata={
            "name": "trackName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="Track",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    musical_part: list[MusicalPart] = field(
        default_factory=list,
        metadata={
            "name": "musicalPart",
            "type": "Element",
            "sequential": True,
        },
    )
    prosody_part: list[ProsodyPart] = field(
        default_factory=list,
        metadata={
            "name": "prosodyPart",
            "type": "Element",
            "sequential": True,
        },
    )


class Vsq3(BaseModel):
    class Meta:
        name = "vsq3"
        namespace = VSQ3_NS

    vender: str | None = field(
        default="Yamaha Corporation",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    version: str | None = field(
        default="3.0.0.0",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    v_voice_table: VVoiceTable = field(
        default_factory=VVoiceTable,
        metadata={
            "name": "vVoiceTable",
            "type": "Element",
            "required": True,
        },
    )
    mixer: Mixer = field(
        default_factory=Mixer,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    master_track: MasterTrack = field(
        default_factory=MasterTrack,
        metadata={
            "name": "masterTrack",
            "type": "Element",
            "required": True,
        },
    )
    vs_track: list[VsTrack] = field(
        default_factory=list,
        metadata={
            "name": "vsTrack",
            "type": "Element",
            "min_occurs": 1,
            "max_occurs": 16,
        },
    )
    mono_track: MonoTrack | None = field(
        default_factory=MonoTrack,
        metadata={
            "name": "seTrack",
            "type": "Element",
            "required": True,
        },
    )
    stereo_track: StereoTrack | None = field(
        default_factory=StereoTrack,
        metadata={
            "name": "karaokeTrack",
            "type": "Element",
            "required": True,
        },
    )
    aux: list[Aux] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    default_singer_name: str = Field(
        "嫣汐",
        title=_("Default Singer Name"),
        description=_("The default singer name to use for all tracks"),
    )



# This submodule is adapted from https://github.com/kislyuk/yq/.
from typing import Any

try:
    import io

    from ruamel.yaml import YAML

    def load_yaml_1_2(text: str) -> dict[str, Any]:
        with YAML(typ="safe") as yml:
            return yml.load(text)

    def dump_yaml_1_2(data: dict[str, Any]) -> str:
        stream = io.StringIO()
        with YAML(output=stream) as yml:
            yml.dump(data)
            return stream.getvalue()

except ImportError:
    import yaml

    from .dumper import get_dumper
    from .loader import get_loader

    def load_yaml_1_2(text: str) -> dict[str, Any]:
        return yaml.load(text, get_loader())

    def dump_yaml_1_2(data: dict[str, Any]) -> str:
        return yaml.dump(
            data,
            None,
            get_dumper(grammar_version="1.2"),
            allow_unicode=True,
            sort_keys=False,
        )



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class OffsetPolicyOption(Enum):
    TIMELINE: Annotated[
        str,
        Field(
            title=_("Act on timeline"),
            description=_(
                'Shift the time axis for each line of lyrics, and keep the "offset" value in the metadata at 0.'
            ),
        ),
    ] = "timeline"
    META: Annotated[
        str,
        Field(
            title=_("Act on metadata"),
            description=_(
                'Write the offset to the metadata "offset" without handling the lyrics time axis. Note that some players may not support the "offset" tag in the metadata, and choosing this option may cause the lyrics to display incorrectly.'
            ),
        ),
    ] = "meta"


class SplitOption(Enum):
    BOTH: Annotated[
        str,
        Field(
            title=_("Both note gap and punctuation"),
            description=_(
                "When the interval between two adjacent notes is greater than or equal to thirty-second note or a punctuation mark is encountered, start a new line."
            ),
        ),
    ] = "both"
    GAP: Annotated[
        str,
        Field(
            title=_("Note gap only"),
            description=_(
                "When the interval between two adjacent notes is greater than or equal to thirty-second note, start a new line."
            ),
        ),
    ] = "gap"
    SYMBOL: Annotated[
        str,
        Field(
            title=_("Punctuation only"),
            description=_("When a punctuation mark is encountered, start a new line."),
        ),
    ] = "symbol"


class OutputOptions(BaseModel):
    artist: str = Field("", title=_("Singer name"))
    title: str = Field("", title=_("Song title"))
    album: str = Field("", title=_("Album name"))
    by: str = Field("", title=_("Lyric editor"))
    offset: int = Field(
        0,
        title=_("Offset"),
        description=_("In milliseconds, positive means ahead, negative means opposite."),
    )
    offset_policy: OffsetPolicyOption = Field(
        title=_("Offset policy"), default=OffsetPolicyOption.TIMELINE
    )
    split_by: SplitOption = Field(title=_("New line by"), default=SplitOption.BOTH)
    ignore_slur_notes: bool = Field(
        title=_("Ignore slur notes"),
        description=_(
            "Ignore '-' lyrics that are used to indicate slur notes in singing synthesizers."
        ),
        default=True,
    )
    timeline: bool = Field(
        title=_("Write timeline"),
        description=_("If you need lyrics without timeline, turn off this option."),
        default=True,
    )
    encoding: str = Field(title=_("Lyric Text encoding"), default="utf-8")



from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class InputOptions(BaseModel):
    pass


class OutputOptions(BaseModel):
    indented: bool = Field(
        default=False,
        title=_("Generate JSON file with indentation"),
        description=_(
            "The indented format is easier to read and modify, but it will take up a larger file size."
        ),
    )



import functools
from collections.abc import Callable

import more_itertools

from libresvip.core.constants import DEFAULT_BPM
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    ParamCurve,
    Params,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)


def _update_curve_points_position(
    curve: ParamCurve,
    func: Callable[[int], float],
    ori_first_bar_ticks: int,
    new_first_bar_ticks: int,
    limit_func: Callable[[int], bool] | None = None,
) -> ParamCurve:
    return ParamCurve(
        points=Points(
            root=[
                point._replace(x=round(func(point.x - ori_first_bar_ticks)) + new_first_bar_ticks)
                for point in curve.points.root
                if limit_func is None or limit_func(point.x - ori_first_bar_ticks)
            ]
        )
    )


def reset_time_axis(project: Project, tempo: float = DEFAULT_BPM) -> Project:
    synchronizer = TimeSynchronizer(
        project.song_tempo_list,
        _is_absolute_time_code=True,
        _default_tempo=tempo,
    )
    new_time_signature = TimeSignature(bar_index=0, numerator=4, denominator=4)
    update_curve_points_position = functools.partial(
        _update_curve_points_position,
        func=synchronizer.get_actual_ticks_from_ticks,
        new_first_bar_ticks=round(new_time_signature.bar_length()),
        ori_first_bar_ticks=round(project.time_signature_list[0].bar_length()),
    )
    new_track_list = []
    for track in project.track_list:
        if isinstance(track, InstrumentalTrack):
            new_track_list.append(
                track.model_copy(
                    deep=True,
                    update={
                        "offset": round(synchronizer.get_actual_ticks_from_ticks(track.offset))
                    },
                )
            )
        elif isinstance(track, SingingTrack):
            new_note_list = []
            for note in track.note_list:
                note_end = round(synchronizer.get_actual_ticks_from_ticks(note.end_pos))
                note_start = round(synchronizer.get_actual_ticks_from_ticks(note.start_pos))
                new_note_list.append(
                    note.model_copy(
                        deep=True,
                        update={
                            "start_pos": note_start,
                            "length": note_end - note_start,
                        },
                    )
                )
            new_track_list.append(
                track.model_copy(
                    deep=True,
                    update={
                        "note_list": new_note_list,
                        "edited_params": Params(
                            pitch=update_curve_points_position(track.edited_params.pitch),
                            volume=update_curve_points_position(track.edited_params.volume),
                            breath=update_curve_points_position(track.edited_params.breath),
                            gender=update_curve_points_position(track.edited_params.gender),
                            strength=update_curve_points_position(track.edited_params.strength),
                        ),
                    },
                )
            )

    return Project(
        song_tempo_list=[SongTempo(bpm=tempo, position=0)],
        time_signature_list=[new_time_signature],
        track_list=new_track_list,
    )


def zoom_project(project: Project, factor: float) -> Project:
    time_signature_list = [
        time_signature.model_copy(update={"bar_index": round(time_signature.bar_index * factor)})
        for time_signature in more_itertools.unique_in_window(
            project.time_signature_list, 2, key=str
        )
    ]
    update_curve_points_position = functools.partial(
        _update_curve_points_position,
        func=factor.__mul__,
        new_first_bar_ticks=round(time_signature_list[0].bar_length()),
        ori_first_bar_ticks=round(project.time_signature_list[0].bar_length()),
    )
    return project.model_copy(
        update={
            "song_tempo_list": [
                song_tempo.model_copy(
                    update={
                        "position": round(song_tempo.position * factor),
                        "bpm": song_tempo.bpm * factor,
                    }
                )
                for song_tempo in project.song_tempo_list
            ],
            "time_signature_list": time_signature_list,
            "track_list": [
                track.model_copy(update={"offset": round(factor * track.offset)})
                if isinstance(track, InstrumentalTrack)
                else track.model_copy(
                    update={
                        "note_list": [
                            note.model_copy(
                                update={
                                    "start_pos": round(factor * note.start_pos),
                                    "length": round(factor * note.length),
                                }
                            )
                            for note in track.note_list
                        ],
                        "edited_params": Params(
                            pitch=update_curve_points_position(track.edited_params.pitch),
                            volume=update_curve_points_position(track.edited_params.volume),
                            breath=update_curve_points_position(track.edited_params.breath),
                            gender=update_curve_points_position(track.edited_params.gender),
                            strength=update_curve_points_position(track.edited_params.strength),
                        ),
                    }
                )
                for track in project.track_list
            ],
        }
    )


def limit_bars(project: Project, max_bars: int) -> Project:
    time_signature_list = [
        time_signature
        for time_signature in project.time_signature_list
        if time_signature.bar_index < max_bars
    ]
    first_bar_length = round(time_signature_list[0].bar_length())
    max_ticks = max_bars * first_bar_length
    update_curve_points_position = functools.partial(
        _update_curve_points_position,
        func=int,
        new_first_bar_ticks=first_bar_length,
        ori_first_bar_ticks=first_bar_length,
        limit_func=max_ticks.__gt__,
    )
    return project.model_copy(
        update={
            "song_tempo_list": [
                song_tempo
                for song_tempo in project.song_tempo_list
                if song_tempo.position < max_ticks
            ],
            "time_signature_list": time_signature_list,
            "track_list": [
                track
                if isinstance(track, InstrumentalTrack)
                else track.model_copy(
                    update={
                        "note_list": [note for note in track.note_list if note.end_pos < max_ticks],
                        "edited_params": Params(
                            pitch=update_curve_points_position(track.edited_params.pitch),
                            volume=update_curve_points_position(track.edited_params.volume),
                            breath=update_curve_points_position(track.edited_params.breath),
                            gender=update_curve_points_position(track.edited_params.gender),
                            strength=update_curve_points_position(track.edited_params.strength),
                        ),
                    }
                )
                for track in project.track_list
            ],
        }
    )



from libresvip.model.point import Point
from libresvip.utils.music_math import (
    interpolate_cosine_ease_in,
    interpolate_cosine_ease_in_out,
    interpolate_cosine_ease_out,
    interpolate_linear,
)

from .model import UTAUPitchBendMode


def interpolate(
    last_point: Point,
    this_point: Point,
    curve_type: UTAUPitchBendMode,
    sampling_interval_tick: int,
) -> list[Point]:
    input_data = [last_point, this_point]
    if curve_type == "s":
        return interpolate_linear(input_data, sampling_interval_tick)
    elif curve_type == "j":
        return interpolate_cosine_ease_in(input_data, sampling_interval_tick)
    elif curve_type == "r":
        return interpolate_cosine_ease_out(input_data, sampling_interval_tick)
    else:
        return interpolate_cosine_ease_in_out(input_data, sampling_interval_tick)



# © 2018 Gerard Marull-Paretas <gerard@teslabs.com>
# © 2014 Mark Harviston <mark.harviston@gmail.com>
# © 2014 Arve Knudsen <arve.knudsen@gmail.com>
# BSD License

"""UNIX specific Quamash functionality."""

from __future__ import annotations

import asyncio
import selectors
from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any

from loguru import logger
from PySide6 import QtCore

from ._common import _fileno

if TYPE_CHECKING:
    from _typeshed import FileDescriptor, FileDescriptorLike


EVENT_READ = 1 << 0
EVENT_WRITE = 1 << 1


class _SelectorMapping(Mapping[int, selectors.SelectorKey]):
    """Mapping of file objects to selector keys."""

    def __init__(self, selector: _Selector) -> None:
        self._selector = selector

    def __len__(self) -> int:
        return len(self._selector._fd_to_key)

    def __getitem__(self, fileobj: FileDescriptorLike) -> selectors.SelectorKey:
        try:
            fd = self._selector._fileobj_lookup(fileobj)
            return self._selector._fd_to_key[fd]
        except KeyError as e:
            msg = f"{fileobj!r} is not registered"
            raise KeyError(msg) from e

    def __iter__(self) -> Iterator[FileDescriptor]:
        yield from iter(self._selector._fd_to_key)


class _Selector(selectors.BaseSelector):
    def __init__(self, parent: _SelectorEventLoop) -> None:
        # this maps file descriptors to keys
        self._fd_to_key: dict[FileDescriptor, selectors.SelectorKey] = {}
        # read-only mapping returned by get_map()
        self.__map = _SelectorMapping(self)
        self.__read_notifiers: dict[FileDescriptorLike, QtCore.QSocketNotifier] = {}
        self.__write_notifiers: dict[FileDescriptorLike, QtCore.QSocketNotifier] = {}
        self.__parent = parent

    def select(self, timeout: float | None = None) -> list[tuple[selectors.SelectorKey, int]]:
        """Implement abstract method even though we don't need it."""
        raise NotImplementedError

    def _fileobj_lookup(self, fileobj: FileDescriptorLike) -> FileDescriptor:
        """Return a file descriptor from a file object.

        This wraps _fileno() to do an exhaustive search in case
        the object is invalid but we still have it in our map.  This
        is used by unregister() so we can unregister an object that
        was previously registered even if it is closed.  It is also
        used by _SelectorMapping.
        """
        try:
            return _fileno(fileobj)
        except ValueError:
            # Do an exhaustive search.
            for key in self._fd_to_key.values():
                if key.fileobj is fileobj:
                    return key.fd
            # Raise ValueError after all.
            raise

    def register(
        self,
        fileobj: FileDescriptorLike,
        events: int,
        data: Any | None = None,
    ) -> selectors.SelectorKey:
        if (not events) or (events & ~(EVENT_READ | EVENT_WRITE)):
            msg = f"Invalid events: {events!r}"
            raise ValueError(msg)

        key = selectors.SelectorKey(fileobj, self._fileobj_lookup(fileobj), events, data)

        if key.fd in self._fd_to_key:
            msg = f"{fileobj!r} (FD {key.fd}) is already registered"
            raise KeyError(msg)

        self._fd_to_key[key.fd] = key

        if events & EVENT_READ:
            notifier = QtCore.QSocketNotifier(key.fd, QtCore.QSocketNotifier.Read)
            notifier.activated["int"].connect(self.__on_read_activated)
            self.__read_notifiers[key.fd] = notifier
        if events & EVENT_WRITE:
            notifier = QtCore.QSocketNotifier(key.fd, QtCore.QSocketNotifier.Write)
            notifier.activated["int"].connect(self.__on_write_activated)
            self.__write_notifiers[key.fd] = notifier

        return key

    def __on_read_activated(self, fd: FileDescriptor) -> None:
        logger.debug("File {} ready to read", fd)
        if key := self._key_from_fd(fd):
            self.__parent._process_event(key, EVENT_READ & key.events)

    def __on_write_activated(self, fd: FileDescriptor) -> None:
        logger.debug("File {} ready to write", fd)
        if key := self._key_from_fd(fd):
            self.__parent._process_event(key, EVENT_WRITE & key.events)

    def unregister(self, fileobj: FileDescriptorLike) -> selectors.SelectorKey:
        def drop_notifier(
            notifiers: dict[FileDescriptorLike, QtCore.QSocketNotifier],
        ) -> None:
            try:
                notifier = notifiers.pop(key.fd)
            except KeyError:
                pass
            else:
                notifier.activated["int"].disconnect()

        try:
            key = self._fd_to_key.pop(self._fileobj_lookup(fileobj))
        except KeyError:
            msg = f"{fileobj!r} is not registered"
            raise KeyError(msg) from None

        drop_notifier(self.__read_notifiers)
        drop_notifier(self.__write_notifiers)

        return key

    def modify(
        self,
        fileobj: FileDescriptorLike,
        events: int,
        data: Any | None = None,
    ) -> selectors.SelectorKey:
        try:
            key = self._fd_to_key[self._fileobj_lookup(fileobj)]
        except KeyError:
            msg = f"{fileobj!r} is not registered"
            raise KeyError(msg) from None
        if events != key.events:
            self.unregister(fileobj)
            key = self.register(fileobj, events, data)
        elif data != key.data:
            # Use a shortcut to update the data.
            key = key._replace(data=data)
            self._fd_to_key[key.fd] = key
        return key

    def close(self) -> None:
        logger.debug("Closing")
        self._fd_to_key.clear()
        self.__read_notifiers.clear()
        self.__write_notifiers.clear()

    def get_map(self) -> Mapping[FileDescriptorLike, selectors.SelectorKey]:
        return self.__map  # type: ignore[return-value]

    def _key_from_fd(self, fd: FileDescriptor) -> selectors.SelectorKey | None:
        """
        Return the key associated to a given file descriptor.

        Parameters:
        fd -- file descriptor

        Returns:
        corresponding key, or None if not found

        """
        try:
            return self._fd_to_key[fd]
        except KeyError:
            return None


class _SelectorEventLoop(asyncio.SelectorEventLoop):
    def __init__(self) -> None:
        selector = _Selector(self)
        super().__init__(selector)

    def _before_run_forever(self) -> None:
        pass

    def _after_run_forever(self) -> None:
        pass

    def _process_event(self, key: selectors.SelectorKey, mask: int) -> None:
        """Selector has delivered us an event."""
        logger.debug("Processing event with key {} and mask {}", key, mask)
        fileobj, (reader, writer) = key.fileobj, key.data
        if mask & selectors.EVENT_READ and reader is not None:
            if reader._cancelled:
                self.remove_reader(fileobj)
            else:
                logger.debug("Invoking reader callback: {}", reader)
                reader._run()
        if mask & selectors.EVENT_WRITE and writer is not None:
            if writer._cancelled:
                self.remove_writer(fileobj)
            else:
                logger.debug("Invoking writer callback: {}", writer)
                writer._run()
